import pytest

from snowflake.core._internal.utils import normalize_and_unquote_name
from tests.integ.external_volume.utils import (
    get_aws_external_volume_template,
    get_aws_external_volume_template_case_sensitive,
    get_azure_external_volume_template,
    get_gcs_external_volume_template,
)


pytestmark = [pytest.mark.internal_only]


def test_list_external_volume(external_volumes, setup_credentials_fixture):
    try:
        external_volume_to_drop = []
        # happy path aws
        external_volume = get_aws_external_volume_template()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]

        external_volume_resources_name = [
            external_volume_resource_temp.name
            for external_volume_resource_temp in external_volumes.iter(like="test_create_aws_external_volume_%")
        ]
        assert external_volume.name.upper() in external_volume_resources_name

        # happy path aws case sensitive
        external_volume = get_aws_external_volume_template_case_sensitive()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]

        external_volume_resources_name = [
            external_volume_resource_temp.name
            for external_volume_resource_temp in external_volumes.iter(like="test_create_aws_external_volume_%")
        ]
        assert normalize_and_unquote_name(external_volume.name) in external_volume_resources_name
        assert len(external_volume_resources_name) >= 2

        # happy path azure
        external_volume = get_azure_external_volume_template()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]

        external_volume_resources_name = [
            external_volume_resource_temp.name
            for external_volume_resource_temp in external_volumes.iter(like="test_create_azure_external_volume_%")
        ]
        assert external_volume.name.upper() in external_volume_resources_name

        # happy path gcs
        external_volume = get_gcs_external_volume_template()
        external_volume_handler = external_volumes.create(external_volume)
        external_volume_to_drop += [external_volume_handler]

        external_volume_resources_name = [
            external_volume_resource_temp.name
            for external_volume_resource_temp in external_volumes.iter(like="test_create_gcs_external_volume_%")
        ]
        assert external_volume.name.upper() in external_volume_resources_name

        # trying to get random volume
        external_volume_resources_name = [
            external_volume_resource_temp.name for external_volume_resource_temp in external_volumes.iter(like="RANDOM")
        ]
        assert len(external_volume_resources_name) == 0

    finally:
        for external_volume_handler in external_volume_to_drop:
            external_volume_handler.drop(if_exists=True)



from unittest import mock

import pytest

from snowflake.core import PollingOperation
from snowflake.core.external_volume import ExternalVolume, ExternalVolumeCollection, ExternalVolumeResource

from ...utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


@pytest.fixture
def external_volumes(fake_root):
    return ExternalVolumeCollection(fake_root)


@pytest.fixture
def external_volume(external_volumes):
    return external_volumes["my_volume"]


def test_create_external_volume(fake_root, external_volumes):
    args = (fake_root, "POST", BASE_URL + "/external-volumes")
    kwargs = extra_params(query_params=[], body={"name": "my_volume", "storage_locations": []})

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        ev_res = external_volumes.create(ExternalVolume(name="my_volume", storage_locations=[]))
        assert isinstance(ev_res, ExternalVolumeResource)
        assert ev_res.name == "my_volume"
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = external_volumes.create_async(ExternalVolume(name="my_volume", storage_locations=[]))
        assert isinstance(op, PollingOperation)
        et_res = op.result()
        assert et_res.name == "my_volume"
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_iter_external_volume(fake_root, external_volumes):
    args = (fake_root, "GET", BASE_URL + "/external-volumes")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        external_volumes.iter()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = external_volumes.iter_async()
        assert isinstance(op, PollingOperation)
        it = op.result()
        assert list(it) == []
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_fetch_external_volume(fake_root, external_volume):
    from snowflake.core.external_volume._generated.models import ExternalVolume as ExternalVolumeModel

    model = ExternalVolumeModel(name="my_volume", storage_locations=[])
    args = (fake_root, "GET", BASE_URL + "/external-volumes/my_volume")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        external_volume.fetch()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response(model.to_json())
        op = external_volume.fetch_async()
        assert isinstance(op, PollingOperation)
        tab = op.result()
        assert tab.to_dict() == ExternalVolume(name="my_volume", storage_locations=[]).to_dict()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_drop_external_volume(fake_root, external_volume):
    args = (fake_root, "DELETE", BASE_URL + "/external-volumes/my_volume")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        external_volume.drop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = external_volume.drop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_undrop_external_volume(fake_root, external_volume):
    args = (fake_root, "POST", BASE_URL + "/external-volumes/my_volume:undrop")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        external_volume.undrop()
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = external_volume.undrop_async()
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)



"""
Cortex Search REST API.

OpenAPI 3.0 specification for the Cortex Search REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr


class Diversity(BaseModel):
    """A model object representing the Diversity resource.

    Constructs an object of type Diversity with the provided properties.

    Parameters
    __________
    group_by : list[str], optional
        Columns to group by.
    max_results : int, optional
        Maximum number of results per group to return.
    """

    group_by: Optional[List[StrictStr]] = None

    max_results: Optional[StrictInt] = None

    __properties = ["group_by", "max_results"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Diversity:
        """Create an instance of Diversity from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> Diversity:
        """Create an instance of Diversity from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return Diversity.model_validate(obj)

        _obj = Diversity.model_validate(
            {
                "group_by": obj.get("group_by"),
                "max_results": obj.get("max_results"),
            }
        )

        return _obj


class DiversityModel:
    def __init__(
        self,  # optional properties
        group_by: Optional[list[str]] = None,
        max_results: Optional[int] = None,
    ):
        """A model object representing the Diversity resource.

        Constructs an object of type Diversity with the provided properties.

        Parameters
        __________
        group_by : list[str], optional
            Columns to group by.
        max_results : int, optional
            Maximum number of results per group to return.
        """
        self.group_by = group_by
        self.max_results = max_results

    __properties = ["group_by", "max_results"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return Diversity(
            group_by=self.group_by,
            max_results=self.max_results,
        )

    @classmethod
    def _from_model(cls, model) -> DiversityModel:
        return DiversityModel(
            group_by=model.group_by,
            max_results=model.max_results,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a Diversity.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the Diversity object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> DiversityModel:
        """Create an instance of Diversity from a dict.

        This method constructs a Diversity object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        Diversity
            A Diversity object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(Diversity.from_dict(obj))


Diversity._model_class = DiversityModel



"""
Cortex Analyst API.

The Snowflake Cortex Analyst API is a REST API that allows end user to chat with their data leveraging semantic models to generate SQL queries.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr


class VerifiedQuery(BaseModel):
    """A model object representing the VerifiedQuery resource.

    Constructs an object of type VerifiedQuery with the provided properties.

    Parameters
    __________
    name : str, optional
        A name for this verified query. Mainly used for display purposes.
    question : str, optional
        The question being answered.
    sql : str, optional
        The correct SQL query for answering the question.
    verified_at : int, optional
        Timestamp at which the query was last verified - measures in seconds since epoch, in UTC.
    verified_by : str, optional
        Name of the person who verified this query.
    """

    name: Optional[StrictStr] = None

    question: Optional[StrictStr] = None

    sql: Optional[StrictStr] = None

    verified_at: Optional[StrictInt] = None

    verified_by: Optional[StrictStr] = None

    __properties = ["name", "question", "sql", "verified_at", "verified_by"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> VerifiedQuery:
        """Create an instance of VerifiedQuery from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> VerifiedQuery:
        """Create an instance of VerifiedQuery from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return VerifiedQuery.model_validate(obj)

        _obj = VerifiedQuery.model_validate(
            {
                "name": obj.get("name"),
                "question": obj.get("question"),
                "sql": obj.get("sql"),
                "verified_at": obj.get("verified_at"),
                "verified_by": obj.get("verified_by"),
            }
        )

        return _obj


class VerifiedQueryModel:
    def __init__(
        self,  # optional properties
        name: Optional[str] = None,
        question: Optional[str] = None,
        sql: Optional[str] = None,
        verified_at: Optional[int] = None,
        verified_by: Optional[str] = None,
    ):
        """A model object representing the VerifiedQuery resource.

        Constructs an object of type VerifiedQuery with the provided properties.

        Parameters
        __________
        name : str, optional
            A name for this verified query. Mainly used for display purposes.
        question : str, optional
            The question being answered.
        sql : str, optional
            The correct SQL query for answering the question.
        verified_at : int, optional
            Timestamp at which the query was last verified - measures in seconds since epoch, in UTC.
        verified_by : str, optional
            Name of the person who verified this query.
        """
        self.name = name
        self.question = question
        self.sql = sql
        self.verified_at = verified_at
        self.verified_by = verified_by

    __properties = ["name", "question", "sql", "verified_at", "verified_by"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return VerifiedQuery(
            name=self.name,
            question=self.question,
            sql=self.sql,
            verified_at=self.verified_at,
            verified_by=self.verified_by,
        )

    @classmethod
    def _from_model(cls, model) -> VerifiedQueryModel:
        return VerifiedQueryModel(
            name=model.name,
            question=model.question,
            sql=model.sql,
            verified_at=model.verified_at,
            verified_by=model.verified_by,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a VerifiedQuery.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the VerifiedQuery object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> VerifiedQueryModel:
        """Create an instance of VerifiedQuery from a dict.

        This method constructs a VerifiedQuery object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        VerifiedQuery
            A VerifiedQuery object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(VerifiedQuery.from_dict(obj))


VerifiedQuery._model_class = VerifiedQueryModel



from typing import Any
from unittest import mock
from unittest.mock import MagicMock

import pytest

from snowflake.core import CreateMode, PollingOperation
from snowflake.core.api_integration import ApiIntegration, ApiIntegrationCollection, ApiIntegrationResource, AwsHook
from snowflake.core.api_integration._generated import TagAssignment, TagReference
from snowflake.core.tag import TagValue

from ..utils import BASE_URL, extra_params, mock_http_response


API_CLIENT_REQUEST = "snowflake.core._generated.api_client.ApiClient.request"


@pytest.fixture
def _mock_collection():
    return MagicMock(database=MagicMock())


@pytest.fixture
def _mock_api_integrations_collection(fake_root):
    return ApiIntegrationCollection(fake_root)


@pytest.fixture()
def api_integrations(fake_root):
    return ApiIntegrationCollection(fake_root)


@pytest.fixture()
def api_integration(api_integrations):
    return api_integrations["my_integration"]


@pytest.fixture
def _mock_api():
    with mock.patch(
        "snowflake.core.api_integration._generated.api.api_integration_api_base.ApiIntegrationApi"
    ) as mock_api:
        yield mock_api.return_value


def parametrize_if_exists():
    return pytest.mark.parametrize("if_exists", [True, False])


def parametrize_mode():
    return pytest.mark.parametrize("mode", [CreateMode.error_if_exists, CreateMode.if_not_exists])


def parametrize_async():
    return pytest.mark.parametrize("is_async", [True, False])


def get_method(resource: Any, name: str, is_async: bool):
    if is_async:
        return getattr(resource, f"{name}_async")
    return getattr(resource, name)


class TestApiIntegrationResource:
    @parametrize_async()
    def test_fetch_api_integration(self, _mock_collection, is_async):
        api_integration = ApiIntegrationResource(name="my_resource", collection=_mock_collection)
        get_method(api_integration, "fetch", is_async)()
        _mock_collection._api.fetch_api_integration.assert_called_once_with("my_resource", async_req=is_async)

    @parametrize_async()
    def test_create_or_alter_api_integration(self, _mock_collection, is_async):
        api_integration = ApiIntegration(
            name="name",
            api_hook=AwsHook(api_provider="AWS_API_GATEWAY", api_aws_role_arn="your_arn", api_key="dummy_api_key"),
            api_allowed_prefixes=["https://snowflake.com"],
            enabled=True,
        )

        resource = ApiIntegrationResource(name="my_resource", collection=_mock_collection)
        get_method(resource, "create_or_alter", is_async)(api_integration)
        _mock_collection._api.create_or_alter_api_integration.assert_called_once_with(
            api_integration.name, api_integration=api_integration, async_req=is_async
        )

    @parametrize_if_exists()
    @parametrize_async()
    def test_drop_api_integration(self, _mock_collection, if_exists, is_async):
        api_integration = ApiIntegrationResource(name="my_resource", collection=_mock_collection)
        get_method(api_integration, "drop", is_async)(if_exists=if_exists)
        _mock_collection._api.delete_api_integration.assert_called_once_with(
            "my_resource", if_exists=if_exists, async_req=is_async
        )


class TestApiIntegrationCollection:
    def test_schema_collection(self, fake_root):
        assert hasattr(fake_root, "api_integrations")

    @parametrize_async()
    def test_iter(self, _mock_api, fake_root, _mock_api_integrations_collection, is_async):
        get_method(_mock_api_integrations_collection, "iter", is_async)(like="%my_resource")

        _mock_api.list_api_integrations.assert_called_once_with(like="%my_resource", async_req=is_async)

    @parametrize_mode()
    @parametrize_async()
    def test_create_api_integration(self, _mock_api, _mock_api_integrations_collection, mode, is_async):
        api_integration = ApiIntegration(
            name="name",
            api_hook=AwsHook(api_provider="AWS_API_GATEWAY", api_aws_role_arn="your_arn", api_key="dummy_api_key"),
            api_allowed_prefixes=["https://snowflake.com"],
            enabled=True,
        )
        get_method(_mock_api_integrations_collection, "create", is_async)(api_integration=api_integration, mode=mode)

        _mock_api.create_api_integration.assert_called_once_with(
            api_integration=api_integration, create_mode=mode, async_req=is_async
        )


def test_set_tags(fake_root, api_integration, tag):
    args = (fake_root, "POST", BASE_URL + "/api-integrations/my_integration:set-tags")
    tags = {tag: TagValue(value="value")}
    kwargs = extra_params(
        body=[
            TagAssignment(
                tag_value=v.value, tag_name=k.name, tag_schema=k.schema.name, tag_database=k.database.name
            ).to_dict()
            for k, v in tags.items()
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        api_integration.set_tags(tags)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = api_integration.set_tags_async(tags)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_unset_tags(fake_root, api_integration, tag):
    args = (fake_root, "POST", BASE_URL + "/api-integrations/my_integration:unset-tags")
    tag_resources = {tag}
    kwargs = extra_params(
        body=[
            TagReference(
                tag_name=tag_res.name, tag_schema=tag_res.schema.name, tag_database=tag_res.database.name
            ).to_dict()
            for tag_res in tag_resources
        ]
    )

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        api_integration.unset_tags(tag_resources)
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        op = api_integration.unset_tags_async(tag_resources)
        assert isinstance(op, PollingOperation)
        op.result()
    mocked_request.assert_called_once_with(*args, **kwargs)


def test_get_tags(fake_root, api_integration):
    args = (fake_root, "GET", BASE_URL + "/api-integrations/my_integration:get-tags")
    kwargs = extra_params()

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        assert api_integration.get_tags() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)

    with mock.patch(API_CLIENT_REQUEST) as mocked_request:
        mocked_request.return_value = mock_http_response()
        op = api_integration.get_tags_async()
        assert isinstance(op, PollingOperation)
        assert op.result() == {}
    mocked_request.assert_called_once_with(*args, **kwargs)



import copy

import pytest

from snowflake.core import CreateMode
from snowflake.core.exceptions import ConflictError
from tests.utils import is_gov_deployment, random_string

from .conftest import (
    test_tag_minimal_template,
    test_tag_template,
)


def test_create_and_fetch(tags, snowflake_region):
    name = random_string(10, "test_tag_create_and_fetch_")
    tag_handle = tags[name]

    try:
        tag = copy.deepcopy(test_tag_template)
        tag.name = name
        tags.create(tag)

        fetched_tag = tag_handle.fetch()

        assert fetched_tag.name == name.upper()
        assert fetched_tag.allowed_values == ["value1", "value2", "value3"]
        assert fetched_tag.comment == "Test tag"
        if is_gov_deployment(snowflake_region):
            assert fetched_tag.propagate == "NONE"
        else:
            assert fetched_tag.propagate is None
        assert fetched_tag.on_conflict is None
        assert fetched_tag.database_name == tags.database.name.upper()
        assert fetched_tag.schema_name == tags.schema.name.upper()
    finally:
        tag_handle.drop(if_exists=True)


def test_create_and_fetch_minimal(tags, snowflake_region):
    name = random_string(10, "test_tag_create_and_fetch_minimal_")
    tag_handle = tags[name]

    try:
        tag = copy.deepcopy(test_tag_minimal_template)
        tag.name = name
        tags.create(tag)

        fetched_tag = tag_handle.fetch()

        assert fetched_tag.name == name.upper()
        assert fetched_tag.allowed_values is None
        assert fetched_tag.comment is None
        if is_gov_deployment(snowflake_region):
            assert fetched_tag.propagate == "NONE"
        else:
            assert fetched_tag.propagate is None
        assert fetched_tag.on_conflict is None
        assert fetched_tag.database_name == tags.database.name.upper()
        assert fetched_tag.schema_name == tags.schema.name.upper()
    finally:
        tag_handle.drop(if_exists=True)


def test_create_and_fetch_create_modes(tags):
    name = random_string(10, "test_tag_create_and_fetch_modes_")
    tag_handle = tags[name]

    try:
        tag = copy.deepcopy(test_tag_template)
        tag.name = name
        tag.comment = "First version"
        tags.create(tag, mode=CreateMode.error_if_exists)
        assert tag_handle.fetch().comment == "First version"

        with pytest.raises(ConflictError):
            tags.create(tag, mode=CreateMode.error_if_exists)

        tag.comment = "Second version"
        tags.create(tag, mode=CreateMode.or_replace)
        assert tag_handle.fetch().comment == "Second version"

        tag.comment = "Should not change"
        tags.create(tag, mode=CreateMode.if_not_exists)
        assert tag_handle.fetch().comment == "Second version"
    finally:
        tag_handle.drop(if_exists=True)



"""
Snowflake Role API.

The Snowflake Role API is a REST API that you can use to access, update, and perform certain actions on Role resource in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictStr


class GrantOn(BaseModel):
    """A model object representing the GrantOn resource.

    Constructs an object of type GrantOn with the provided properties.

    Parameters
    __________
    created_on : datetime, optional
        Date and time when the grant was created — **Read-only:** *any user-provided value will be ignored.*
    privilege : str, optional
        The name of the privilege — **Read-only:** *any user-provided value will be ignored.*
    granted_on : str, optional
        The type of of the role — **Read-only:** *any user-provided value will be ignored.*
    name : str, optional
        The name of the role — **Read-only:** *any user-provided value will be ignored.*
    granted_to : str, optional
        The type of the grantee — **Read-only:** *any user-provided value will be ignored.*
    grantee_name : str, optional
        The name of the grantee — **Read-only:** *any user-provided value will be ignored.*
    grant_option : str, optional
        If true, allows the recipient role to grant the privileges to other roles — **Read-only:** *any user-provided value will be ignored.*
    granted_by : str, optional
        The role that granted this privilege to this grantee — **Read-only:** *any user-provided value will be ignored.*
    granted_by_role_type : str, optional
        Type of the role that granted this privilege to this grantee — **Read-only:** *any user-provided value will be ignored.*
    """

    created_on: Optional[datetime] = None

    privilege: Optional[StrictStr] = None

    granted_on: Optional[StrictStr] = None

    name: Optional[StrictStr] = None

    granted_to: Optional[StrictStr] = None

    grantee_name: Optional[StrictStr] = None

    grant_option: Optional[StrictStr] = None

    granted_by: Optional[StrictStr] = None

    granted_by_role_type: Optional[StrictStr] = None

    __properties = [
        "created_on",
        "privilege",
        "granted_on",
        "name",
        "granted_to",
        "grantee_name",
        "grant_option",
        "granted_by",
        "granted_by_role_type",
    ]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> GrantOn:
        """Create an instance of GrantOn from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update(
                {
                    "created_on",
                    "privilege",
                    "granted_on",
                    "name",
                    "granted_to",
                    "grantee_name",
                    "grant_option",
                    "granted_by",
                    "granted_by_role_type",
                }
            )

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> GrantOn:
        """Create an instance of GrantOn from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return GrantOn.model_validate(obj)

        _obj = GrantOn.model_validate(
            {
                "created_on": obj.get("created_on"),
                "privilege": obj.get("privilege"),
                "granted_on": obj.get("granted_on"),
                "name": obj.get("name"),
                "granted_to": obj.get("granted_to"),
                "grantee_name": obj.get("grantee_name"),
                "grant_option": obj.get("grant_option"),
                "granted_by": obj.get("granted_by"),
                "granted_by_role_type": obj.get("granted_by_role_type"),
            }
        )

        return _obj


class GrantOnModel:
    def __init__(
        self,  # optional properties
        created_on: Optional[datetime] = None,
        privilege: Optional[str] = None,
        granted_on: Optional[str] = None,
        name: Optional[str] = None,
        granted_to: Optional[str] = None,
        grantee_name: Optional[str] = None,
        grant_option: Optional[str] = None,
        granted_by: Optional[str] = None,
        granted_by_role_type: Optional[str] = None,
    ):
        """A model object representing the GrantOn resource.

        Constructs an object of type GrantOn with the provided properties.

        Parameters
        __________
        created_on : datetime, optional
            Date and time when the grant was created
        privilege : str, optional
            The name of the privilege
        granted_on : str, optional
            The type of of the role
        name : str, optional
            The name of the role
        granted_to : str, optional
            The type of the grantee
        grantee_name : str, optional
            The name of the grantee
        grant_option : str, optional
            If true, allows the recipient role to grant the privileges to other roles.
        granted_by : str, optional
            The role that granted this privilege to this grantee
        granted_by_role_type : str, optional
            Type of the role that granted this privilege to this grantee
        """
        self.created_on = created_on
        self.privilege = privilege
        self.granted_on = granted_on
        self.name = name
        self.granted_to = granted_to
        self.grantee_name = grantee_name
        self.grant_option = grant_option
        self.granted_by = granted_by
        self.granted_by_role_type = granted_by_role_type

    __properties = [
        "created_on",
        "privilege",
        "granted_on",
        "name",
        "granted_to",
        "grantee_name",
        "grant_option",
        "granted_by",
        "granted_by_role_type",
    ]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return GrantOn(
            created_on=self.created_on,
            privilege=self.privilege,
            granted_on=self.granted_on,
            name=self.name,
            granted_to=self.granted_to,
            grantee_name=self.grantee_name,
            grant_option=self.grant_option,
            granted_by=self.granted_by,
            granted_by_role_type=self.granted_by_role_type,
        )

    @classmethod
    def _from_model(cls, model) -> GrantOnModel:
        return GrantOnModel(
            created_on=model.created_on,
            privilege=model.privilege,
            granted_on=model.granted_on,
            name=model.name,
            granted_to=model.granted_to,
            grantee_name=model.grantee_name,
            grant_option=model.grant_option,
            granted_by=model.granted_by,
            granted_by_role_type=model.granted_by_role_type,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a GrantOn.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the GrantOn object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> GrantOnModel:
        """Create an instance of GrantOn from a dict.

        This method constructs a GrantOn object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        GrantOn
            A GrantOn object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(GrantOn.from_dict(obj))


GrantOn._model_class = GrantOnModel



"""
Common generated classes shared across all resources.

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

import json
import logging
import re
import sys

import urllib3

from snowflake.core._http_requests import create_connection_pool
from snowflake.core.exceptions import (
    APIError,
    ConflictError,
    ForbiddenError,
    NotFoundError,
    ServerError,
    UnauthorizedError,
    _APIValueError,
)
from snowflake.core.rest import RESTResponse


# Until we work figure out how to have the generated code use APIError
#  instead of ApiException we alias it.
ApiException = APIError

logger = logging.getLogger(__name__)

RETRY_STATUS_CODES: set[int] = {202, 429, 503, 504}


def evaluate_status_codes(root, r):
    retry_request = root.parameters().should_retry_request and r.status in RETRY_STATUS_CODES
    sys.tracebacklimit = 0
    if root.parameters().should_print_verbose_stack_trace:
        sys.tracebacklimit = None
    if not 200 <= r.status <= 299 and not retry_request:
        if r.status == 401:
            raise UnauthorizedError(http_resp=r, root=root)

        if r.status == 403:
            raise ForbiddenError(http_resp=r, root=root)

        if r.status == 404:
            raise NotFoundError(http_resp=r, root=root)

        if r.status == 409:
            raise ConflictError(http_resp=r, root=root)

        if 500 <= r.status <= 599:
            raise ServerError(http_resp=r, root=root)

        raise APIError(http_resp=r, root=root)

    return r


class RESTClientObject:
    def __init__(self, *args, **kwargs):
        self.pool_manager = create_connection_pool(
            *args,
            **kwargs,
        )

    def request(
        self,
        root,
        method,
        url,
        query_params=None,
        headers=None,
        body=None,
        post_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        """Perform requests.

        :param method: http request method
        :param url: http request url
        :param query_params: query parameters in the url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]

        if post_params and body:
            raise _APIValueError("body parameter cannot be used with post_params parameter.")

        post_params = post_params or {}
        headers = headers or {}

        timeout = None
        if _request_timeout:
            if isinstance(_request_timeout, (int, float)):
                timeout = urllib3.Timeout(total=_request_timeout)
            elif isinstance(_request_timeout, tuple) and len(_request_timeout) == 2:
                timeout = urllib3.Timeout(connect=_request_timeout[0], read=_request_timeout[1])

        try:
            # For `POST`, `PUT`, `PATCH`, `OPTIONS`, `DELETE`
            if method in ["POST", "PUT", "PATCH", "OPTIONS", "DELETE"]:
                # no content type provided or payload is json
                if not headers.get("Content-Type") or re.search("json", headers["Content-Type"], re.IGNORECASE):
                    request_body = None
                    if body is not None:
                        request_body = json.dumps(body)
                    r = self.pool_manager.request(
                        root,
                        method,
                        url,
                        body=request_body,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers,
                    )
                elif headers["Content-Type"] == "application/x-www-form-urlencoded":
                    r = self.pool_manager.request(
                        root,
                        method,
                        url,
                        fields=post_params,
                        encode_multipart=False,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers,
                    )
                elif headers["Content-Type"] == "multipart/form-data":
                    # must del headers['Content-Type'], or the correct
                    # Content-Type which generated by urllib3 will be
                    # overwritten.
                    del headers["Content-Type"]
                    r = self.pool_manager.request(
                        root,
                        method,
                        url,
                        fields=post_params,
                        encode_multipart=True,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers,
                    )
                # Pass a `string` parameter directly in the body to support
                # other content types than Json when `body` argument is
                # provided in serialized form
                elif isinstance(body, str) or isinstance(body, bytes):
                    request_body = body
                    r = self.pool_manager.request(
                        root,
                        method,
                        url,
                        body=request_body,
                        preload_content=_preload_content,
                        timeout=timeout,
                        headers=headers,
                    )
                else:
                    # Cannot generate the request from given parameters
                    msg = """Cannot prepare a request message for provided
                             arguments. Please check that your arguments match
                             declared content type."""
                    raise APIError(status=0, reason=msg, root=root)
            # For `GET`, `HEAD`
            else:
                r = self.pool_manager.request(
                    root,
                    method,
                    url,
                    fields={},
                    preload_content=_preload_content,
                    timeout=timeout,
                    headers=headers,
                )
        except urllib3.exceptions.SSLError as e:
            msg = f"{type(e).__name__}\n{str(e)}"
            raise APIError(status=0, reason=msg, root=root)

        if _preload_content:
            r = RESTResponse(r)

            # log response body
            logger.debug("response body: %s", r.data)

        return evaluate_status_codes(root, r)

    def get_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "GET",
            url,
            headers=headers,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            query_params=query_params,
        )

    def head_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "HEAD",
            url,
            headers=headers,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            query_params=query_params,
        )

    def options_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "OPTIONS",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def delete_request(
        self, root, url, headers=None, query_params=None, body=None, _preload_content=True, _request_timeout=None
    ):
        return self.request(
            root,
            "DELETE",
            url,
            headers=headers,
            query_params=query_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def post_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "POST",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def put_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "PUT",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def patch_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "PATCH",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )


class StoredProcClientObject:
    @staticmethod
    def _is_correct_xp_request(payload) -> bool:
        return "data" in payload and "code" in payload and "message" in payload and "success" in payload

    def request(
        self,
        root,
        method,
        url,
        query_params={},  # noqa:B006
        headers={},  # noqa:B006
        body={},  # noqa:B006
        post_params={},  # noqa:B006
        _preload_content=True,
        _request_timeout=None,
    ):
        """Perform requests.

        :param method: http request method
        :param url: http request url
        :param query_params: query parameters in the url
        :param headers: http request headers
        :param body: request json body, for `application/json`
        :param post_params: request post parameters,
                            `application/x-www-form-urlencoded`
                            and `multipart/form-data`
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        """
        method = method.upper()
        assert method in ["GET", "HEAD", "DELETE", "POST", "PUT", "PATCH", "OPTIONS"]
        import _snowflake

        parsed_url = urllib3.util.parse_url(url)
        query_params = query_params or {}
        response_dict = _snowflake.send_snow_api_request(
            method, parsed_url.path, dict(query_params), headers, body, post_params, _request_timeout
        )
        json_content = json.loads(response_dict["content"])
        if self._is_correct_xp_request(json_content):
            r = urllib3.HTTPResponse(body=json.dumps(json_content["data"]))
        else:
            r = urllib3.HTTPResponse(body=json.dumps(json_content).encode("utf-8"))
        r.status = response_dict["status"]
        formatted_headers = {key.strip(): value for key, value in response_dict["headers"].items()}
        r.headers = formatted_headers
        logger.info(
            "Received an HTTP response of status %d",
            r.status,
        )
        if _preload_content:
            r = RESTResponse(r)
            # log response body
            logger.debug("response body: %s", r.data)

        return evaluate_status_codes(root, r)

    def get_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "GET",
            url,
            headers=headers,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            query_params=query_params,
        )

    def head_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "HEAD",
            url,
            headers=headers,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            query_params=query_params,
        )

    def options_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "OPTIONS",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def delete_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "DELETE",
            url,
            headers=headers,
            query_params=query_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def post_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "POST",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def put_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "PUT",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )

    def patch_request(
        self,
        root,
        url,
        headers=None,
        query_params=None,
        post_params=None,
        body=None,
        _preload_content=True,
        _request_timeout=None,
    ):
        return self.request(
            root,
            "PATCH",
            url,
            headers=headers,
            query_params=query_params,
            post_params=post_params,
            _preload_content=_preload_content,
            _request_timeout=_request_timeout,
            body=body,
        )



from typing import TYPE_CHECKING, Generic, Protocol, TypeVar

from snowflake.connector import SnowflakeConnection
from snowflake.core._common.base_collection import ObjectCollection, ObjectReferenceMixin


if TYPE_CHECKING:
    from snowflake.core import Root
    from snowflake.core.database import DatabaseResource

T = TypeVar("T")


class DatabaseObjectCollectionParent(ObjectCollection[T], Generic[T]):
    def __init__(self, database: "DatabaseResource", ref_class: type[T]) -> None:
        super().__init__(ref_class)
        self._database = database

    @property
    def _connection(self) -> SnowflakeConnection:
        return self._database._connection

    @property
    def database(self) -> "DatabaseResource":
        """The DatabaseResource this collection belongs to."""
        return self._database

    @property
    def root(self) -> "Root":
        """The Root object this collection belongs to."""
        return self.database.collection.root

    def __repr__(self) -> str:
        type_name = type(self).__name__
        return f"<{type_name}: {self.database.name!r}>"


class DatabaseObjectReferenceProtocol(Protocol[T]):
    @property
    def collection(self) -> DatabaseObjectCollectionParent[T]: ...

    @property
    def name(self) -> str: ...

    @property
    def database(self) -> "DatabaseResource": ...


class DatabaseObjectReferenceMixin(Generic[T], ObjectReferenceMixin[DatabaseObjectCollectionParent[T]]):
    @property
    def database(self: DatabaseObjectReferenceProtocol[T]) -> "DatabaseResource":
        """The DatabaseResource this reference belongs to."""
        return self.collection.database

    @property
    def qualified_name(self: DatabaseObjectReferenceProtocol[T]) -> str:
        """Return the qualified name of the object this reference points to."""
        return f"{self.database.name}.{self.name}"

    def __repr__(self: DatabaseObjectReferenceProtocol[T]) -> str:
        type_name = type(self).__name__
        qualified_name = f"{self.database.name}.{self.name}"
        return f"<{type_name}: {qualified_name!r}>"



"""
Cortex Inference API.

OpenAPI 3.0 specification for the Cortex REST API  # noqa: E501

The version of the OpenAPI document: 0.1.0
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict

from snowflake.core.cortex.inference_service._generated.models.open_ai_reasoning import (
    OpenAIReasoning,
    OpenAIReasoningModel,
)


class OpenAIOutput(BaseModel):
    """A model object representing the OpenAIOutput resource.

    Constructs an object of type OpenAIOutput with the provided properties.

    Parameters
    __________
    reasoning : OpenAIReasoning, optional
    """

    reasoning: Optional[OpenAIReasoning] = None

    __properties = ["reasoning"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> OpenAIOutput:
        """Create an instance of OpenAIOutput from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of reasoning
        if self.reasoning:
            _dict["reasoning"] = self.reasoning.to_dict()

        # set to None if reasoning (nullable) is None
        if self.reasoning is None:
            _dict["reasoning"] = None

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> OpenAIOutput:
        """Create an instance of OpenAIOutput from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return OpenAIOutput.model_validate(obj)

        _obj = OpenAIOutput.model_validate(
            {
                "reasoning": OpenAIReasoning.from_dict(obj.get("reasoning"))
                if obj.get("reasoning") is not None
                else None,
            }
        )

        return _obj


class OpenAIOutputModel:
    def __init__(
        self,  # optional properties
        reasoning: Optional[OpenAIReasoning] = None,
    ):
        """A model object representing the OpenAIOutput resource.

        Constructs an object of type OpenAIOutput with the provided properties.

        Parameters
        __________
        reasoning : OpenAIReasoning, optional
        """
        self.reasoning = reasoning

    __properties = ["reasoning"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return OpenAIOutput(
            reasoning=self.reasoning._to_model() if self.reasoning is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> OpenAIOutputModel:
        return OpenAIOutputModel(
            reasoning=OpenAIReasoningModel._from_model(model.reasoning) if model.reasoning else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a OpenAIOutput.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the OpenAIOutput object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> OpenAIOutputModel:
        """Create an instance of OpenAIOutput from a dict.

        This method constructs a OpenAIOutput object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        OpenAIOutput
            A OpenAIOutput object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(OpenAIOutput.from_dict(obj))


OpenAIOutput._model_class = OpenAIOutputModel



import time
import warnings

from contextlib import ExitStack
from datetime import timedelta

import pytest

from snowflake.core._common import CreateMode
from snowflake.core._internal.utils import normalize_name
from snowflake.core.exceptions import APIError
from snowflake.core.task import StoredProcedureCall
from snowflake.core.task.context import TaskContext
from snowflake.core.task.dagv1 import DAG, DAGOperation, DAGTask, DAGTaskBranch, _use_func_return_value
from snowflake.core.warehouse import Warehouse

from ...utils import get_task_history, random_object_name


def test_deploy_dag(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task2 << task1 >> task3
    op = DAGOperation(schema)
    op.deploy(dag, mode=CreateMode.or_replace)
    try:
        fetched = schema.tasks[test_dag].fetch_task_dependents()
        fetched = sorted(fetched, key=lambda x: x.name.lower())
        assert len(fetched) == 4
        assert fetched[0].name.lower() == test_dag
        assert fetched[1].name.lower() == f"{test_dag}$task1"
        assert fetched[2].name.lower() == f"{test_dag}$task2"
        assert fetched[3].name.lower() == f"{test_dag}$task3"
    finally:
        op.drop(dag)


def test_deploy_dag_with_finalizer(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        DAGTask("task3", "select 3", is_finalizer=True)
        task2 << task1
    op = DAGOperation(schema)
    op.deploy(dag, mode=CreateMode.or_replace)
    try:
        fetched = schema.tasks[test_dag].fetch_task_dependents()
        assert len(fetched) == 4
        assert fetched[0].name.lower() == test_dag
        assert fetched[1].name.lower() == f"{test_dag}$task1"
    finally:
        op.drop(dag)


def test_deploy_dag_default_mode(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task2 << task1 >> task3
    op = DAGOperation(schema)
    # Deploy with the default mode DeploymentMode.error_if_exists for coverage.
    op.deploy(dag)
    try:
        fetched = schema.tasks[test_dag].fetch_task_dependents()
        assert len(fetched) == 4
        assert fetched[0].name.lower() == test_dag
        assert fetched[1].name.lower() == f"{test_dag}$task1"
    finally:
        op.drop(dag)


@pytest.mark.usefixtures("backup_database_schema")
def test_deploy_dag_to_not_default_schema(root, database):
    database = database.name
    new_schema_name = random_object_name()
    test_dag = random_object_name()
    root.connection.execute_string(f"create schema {database}.{new_schema_name}")
    try:
        new_schema = root.databases[database].schemas[new_schema_name]
        with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
            task1 = DAGTask("task1", "select 1")
            task2 = DAGTask("task2", "select 2")
            task3 = DAGTask("task3", "select 3")
            task2 << task1 >> task3
        op = DAGOperation(new_schema)
        op.deploy(dag, mode=CreateMode.or_replace)
        fetched = new_schema.tasks[test_dag].fetch_task_dependents()
        assert len(fetched) == 4
        assert fetched[0].name.lower() == test_dag
        assert fetched[1].name.lower() == f"{test_dag}$task1"
        op.drop(dag)
    finally:
        root.connection.execute_string(f"drop schema if exists {new_schema_name}")


def test_deploy_dag_circle(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task1 >> task2 >> task3 >> task1
    op = DAGOperation(schema)
    with pytest.raises(ValueError) as ex:
        op.deploy(dag, mode=CreateMode.or_replace)
    assert ex.match("There is a cycle in the task graph.")


def test_deploy_dag_5_layers(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        prev_task = DAGTask("task1", "select 1")
        for i in range(2, 6):
            task = DAGTask(f"task{i}", f"select {i}")
            prev_task >> task
            prev_task = task
    op = DAGOperation(schema)
    op.deploy(dag, mode=CreateMode.or_replace)
    try:
        fetched = schema.tasks[test_dag].fetch_task_dependents()
        assert len(fetched) == 6
        assert [x.name.lower() for x in fetched] == [test_dag, *[f"{test_dag}$task{i + 1}" for i in range(5)]]
    finally:
        op.drop(dag)


@pytest.mark.snowpark
@pytest.mark.usefixtures("anaconda_package_available")
def test_deploy_dag_definition_python(schema, db_parameters):
    from snowflake.snowpark import Session

    def foo(session: Session) -> None:
        session.sql("select 1").collect()

    test_stage = random_object_name()
    test_dag = random_object_name()
    schema._connection.execute_string(f"create or replace stage {test_stage}")
    try:
        with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
            task1 = DAGTask(
                "task1",
                StoredProcedureCall(foo, stage_location=test_stage, packages=["snowflake-snowpark-python"]),
                warehouse=db_parameters["warehouse"],
            )
            task2 = DAGTask("task2", "select 2")
            task1 >> task2
        op = DAGOperation(schema)
        op.deploy(dag, mode=CreateMode.or_replace)
        try:
            fetched = schema.tasks[task1.full_name].fetch()
            assert "foo" in fetched.definition
        finally:
            op.drop(dag)
    finally:
        schema._connection.execute_string(f"drop stage {test_stage}")


@pytest.mark.snowpark
@pytest.mark.usefixtures("anaconda_package_available")
def test_deploy_dag_python_function_directly(schema, db_parameters):
    from snowflake.snowpark import Session

    def foo0(session: Session) -> None:
        session.sql("select 'foo0'").collect()

    def foo1(session: Session) -> None:
        session.sql("select 'foo1'").collect()

    def foo2(session: Session) -> None:
        session.sql("select 'foo2'").collect()

    def foo3(session: Session) -> None:
        session.sql("select 'foo3'").collect()

    def foo4(session: Session) -> None:
        session.sql("select 'foo4'").collect()

    def foo5(session: Session) -> None:
        session.sql("select 'foo5'").collect()

    def foo6(session: Session) -> None:
        session.sql("select 'foo6'").collect()

    test_stage = random_object_name()
    test_dag = random_object_name()
    schema._connection.execute_string(f"create or replace stage {test_stage}")
    try:
        with DAG(
            test_dag,
            schedule=timedelta(minutes=10),
            stage_location=test_stage,
            warehouse=db_parameters["warehouse"],
            packages=["snowflake-snowpark-python"],
        ) as dag:
            task0 = DAGTask("foo0", foo0, is_serverless=True)
            task0 >> foo1 >> [foo3, foo4]
            task0 << foo2 << [foo5, foo6]

        op = DAGOperation(schema)
        op.deploy(dag, mode=CreateMode.or_replace)
        try:
            assert "foo0" in schema.tasks[task0.full_name].fetch().definition
            assert "foo1" in schema.tasks[f"{test_dag}$foo1"].fetch().definition
            assert "foo2" in schema.tasks[f"{test_dag}$foo2"].fetch().definition
            assert "foo2" in schema.tasks[f"{test_dag}$foo2"].fetch().definition
            assert "foo3" in schema.tasks[f"{test_dag}$foo3"].fetch().definition
            assert "foo4" in schema.tasks[f"{test_dag}$foo4"].fetch().definition
            assert "foo5" in schema.tasks[f"{test_dag}$foo5"].fetch().definition
            assert "foo6" in schema.tasks[f"{test_dag}$foo6"].fetch().definition
            dependents = schema.tasks[test_dag].fetch_task_dependents()
            assert dependents[0].name.upper() == test_dag.upper()
            # foo5 and foo6 are next to dependents[0].
            assert {dependents[1].name.upper(), dependents[2].name.upper()} == {
                f"{test_dag}$foo5".upper(),
                f"{test_dag}$foo6".upper(),
            }
            assert dependents[3].name.upper() == f"{test_dag}$foo2".upper()
            assert dependents[4].name.upper() == f"{test_dag}$foo0".upper()
            assert dependents[5].name.upper() == f"{test_dag}$foo1".upper()
            assert {dependents[6].name.upper(), dependents[7].name.upper()} == {
                f"{test_dag}$foo3".upper(),
                f"{test_dag}$foo4".upper(),
            }
        finally:
            op.drop(dag)
    finally:
        schema._connection.execute_string(f"drop stage {test_stage}")


@pytest.mark.snowpark
@pytest.mark.usefixtures("anaconda_package_available")
@pytest.mark.parametrize("reverse", [True, False])
def test_deploy_dag_with_branch(schema, db_parameters, reverse):
    from snowflake.snowpark import Session

    def task_handler(session: Session) -> None:
        pass  # do nothing

    def task_branch_handler(session: Session) -> None:
        tc = TaskContext(session)
        tc.set_return_value("task3")

    def task4_handler(session: Session) -> None:
        tc = TaskContext(session)
        tc.set_return_value("task4")

    test_stage = random_object_name()
    test_dag = random_object_name()
    schema._connection.execute_string(f"create or replace stage {test_stage}")
    try:
        with DAG(
            test_dag,
            schedule=timedelta(minutes=10),
            stage_location=test_stage,
            packages=["snowflake-snowpark-python"],
            warehouse=db_parameters["warehouse"],
        ) as dag:
            task1 = DAGTask("task1", task_handler, warehouse=db_parameters["warehouse"])
            task1_branch = DAGTaskBranch("task1_branch", task_branch_handler, warehouse=db_parameters["warehouse"])
            task2 = DAGTask("task2", task_handler, is_serverless=True)
            task3 = DAGTask("task3", task_handler, warehouse=db_parameters["warehouse"], condition="1=1")
            if not reverse:
                task1 >> task1_branch >> [task2, task3]
            else:
                task1_branch << task1
                task1_branch >> [task2, task3]
            task2_branch = DAGTaskBranch("task2_branch", task_branch_handler, warehouse=db_parameters["warehouse"])
            if not reverse:
                task2 >> task2_branch >> task4_handler
            else:
                DAGTask("task4_handler", task4_handler) << task2_branch << task2
        op = DAGOperation(schema)
        op.deploy(dag, mode=CreateMode.or_replace)
        try:
            dependents = schema.tasks[dag.name].fetch_task_dependents()
            task4 = dependents[-1]
            assert len(dependents) == 7
            assert (
                f"SYSTEM$GET_PREDECESSOR_RETURN_VALUE('{normalize_name(task1_branch.full_name)}') = '{task2.name}'"
                == task2.condition
            )
            assert (
                f"SYSTEM$GET_PREDECESSOR_RETURN_VALUE('{normalize_name(task1_branch.full_name)}') = "
                f"'{task3.name}' and 1=1" == task3.condition
            )
            assert (
                f"SYSTEM$GET_PREDECESSOR_RETURN_VALUE('{normalize_name(task2_branch.full_name)}') = 'task4_handler'"
                == task4.condition
            )
            assert task1_branch in task2.predecessors
            assert task1_branch in task3.predecessors
            fqal_task2_name = f"{schema.database.name}.{schema.name}.{task2_branch.full_name}".upper()
            assert fqal_task2_name in task4.predecessors or task2_branch.full_name.upper() in task4.predecessors
        finally:
            op.drop(dag)
    finally:
        schema._connection.execute_string(f"drop stage {test_stage}")


@pytest.mark.snowpark
@pytest.mark.usefixtures("anaconda_package_available")
def test_dag_use_function_return_value(schema, db_parameters):
    from snowflake.snowpark import Session

    test_stage = f"{schema.database.name}.{schema.name}.{random_object_name()}"
    test_dag = random_object_name()
    schema._connection.execute_string(f"create or replace stage {test_stage}")

    def task_handler(session: Session) -> str:
        return "task1_return_value"

    try:
        with DAG(
            test_dag,
            schedule=timedelta(minutes=10),
            stage_location=test_stage,
            packages=["snowflake-snowpark-python"],
            use_func_return_value=True,
        ) as dag:
            task1 = DAGTask("task1", task_handler, warehouse=db_parameters["warehouse"])
        op = DAGOperation(schema)
        op.deploy(dag, mode=CreateMode.or_replace)
        try:
            op.run(dag)
            # get the history
            time.sleep(1)
            time_count = 0
            time_limit = 120
            task_name = task1.full_name
            while True:
                result = get_task_history(schema.root.session, task_name)
                if len(result) == 0 or (state := result[0]["STATE"]) in ["SCHEDULED", "EXECUTING"]:
                    if time_count > time_limit:  # run for 2 min
                        raise ValueError(f"Running more than {time_limit} seconds. task: {task_name}, result: {result}")
                    time_count += 10
                    time.sleep(10)
                    continue
                elif state == "SUCCEEDED":
                    assert result[0]["RETURN_VALUE"] == "task1_return_value"
                    break
                else:
                    raise ValueError(f"unexpected state: {state}. {result}")
        finally:
            op.drop(dag)
    finally:
        schema._connection.execute_string(f"drop stage {test_stage}")


def test__use_func_return_value(session):
    from snowflake.snowpark import Session

    def foo(session: Session) -> str:
        return "abc"

    wrapped_foo = _use_func_return_value(foo)
    assert wrapped_foo(session) == "abc"


def test_drop_dag_not_exist(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        prev_task = DAGTask("task1", "select 1")
        for i in range(2, 6):
            task = DAGTask(f"task{i}", f"select {i}")
            task << prev_task
            prev_task = task
    op = DAGOperation(schema)
    op.drop(dag)


@pytest.mark.parametrize("drop_finalizer", [True, False])
def test_drop_dag_by_name(schema, drop_finalizer):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task = DAGTask("task1", "select 1")
        finalizer = DAGTask("finalizer", "select 2", is_finalizer=True)
        op = DAGOperation(schema)
        op.deploy(dag)
        op.drop(dag.name, drop_finalizer=drop_finalizer)  # expect no exception thrown
    task_list = [t.name for t in schema.tasks.iter()]
    assert task.full_name.upper() not in task_list
    if drop_finalizer:
        assert finalizer.full_name.upper() not in task_list
    else:
        assert finalizer.full_name.upper() in task_list


def test_drop_dag_without_remaining_task(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task2 >> task1
    op = DAGOperation(schema)
    op.deploy(dag)
    op.run(dag)

    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task3", "select 1")
        task2 = DAGTask("task2", "select 2")
        task2 >> task1
    op.deploy(dag, mode=CreateMode.or_replace)
    op.run(dag)
    task_list = [i.name for i in schema.tasks.iter()]
    assert "task1" not in task_list
    op.drop(dag)


def test_run_dag(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task2 >> task1 << task3
    op = DAGOperation(schema)
    op.deploy(dag, mode=CreateMode.or_replace)
    try:
        current_run1 = op.get_current_dag_runs(dag)
        assert len(current_run1) == 1
        assert current_run1[0].state == "SCHEDULED"
        op.run(dag)
        current_run2 = op.get_current_dag_runs(dag)
        assert len(current_run2) == 1
        assert current_run2[0].scheduled_time < current_run1[0].scheduled_time
        op.get_complete_dag_runs(
            dag, error_only=False
        )  # TODO: There is a latency so no assertion now. Review this later.
    finally:
        op.drop(dag)


def test_run_retry_last(schema):
    # This test will not retry a task successfully. Instead, it asserts that retry failure error message is expected.
    # So this test makes sure the client side has sent the right request to the server.
    #
    # To test a retry from a failed task, it needs to wait some time for the task to fail.
    # It would be time-consuming and flaky.
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        DAGTask("task1", "select 1")
    op = DAGOperation(schema)
    op.deploy(dag, mode=CreateMode.or_replace)
    try:
        with pytest.raises(APIError) as exec_info:
            op.run(dag, retry_last=True)
        assert exec_info.value.status == 400
        assert exec_info.match("(?i)Cannot perform retry")
    finally:
        op.drop(dag)


def test_iter_dags(schema):
    random_name = random_object_name()
    test_dag1 = random_name + "a"
    test_dag2 = random_name + "b"
    with DAG(test_dag1, schedule=timedelta(minutes=10)) as dag1:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task2 << task1 >> task3
    with DAG(test_dag2, schedule=timedelta(minutes=10)) as dag2:
        task4 = DAGTask("task4", "select 4")
        task5 = DAGTask("task5", "select 5")
        task6 = DAGTask("task6", "select 6")
        task5 << task4 >> task6
    op = DAGOperation(schema)
    try:
        op.deploy(dag1, mode=CreateMode.or_replace)
        try:
            op.deploy(dag2, mode=CreateMode.or_replace)
            fetched_dags = list(op.iter_dags(like=random_name + "%"))
            assert len(fetched_dags) == 2
            assert set(fetched_dags) == {test_dag1.upper(), test_dag2.upper()}
        finally:
            op.drop(dag2)
    finally:
        op.drop(dag1)


def test_dagtask_without_dag_context(schema):
    with pytest.raises(ValueError):
        DAGTask("task1", "select 1")


def test_get_task(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")  # noqa: F841
        task2 = DAGTask("task2", "select 2")
        assert dag.get_task("task2") == task2
        assert dag.get_task("notask") is None


def test_get_finalizer(schema):
    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1", is_finalizer=True)
        task2 = DAGTask("task2", "select 2")
        assert dag.get_task("task2") == task2
        assert dag.get_task("task1") == task1
        assert dag.get_finalizer_task() == task1
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        assert dag.get_task("task1") == task1
        assert dag.get_finalizer_task() is None


def test_add_one_predecessor(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task1.add_predecessors(task2)
        assert task1.predecessors == {task2}
        assert dag.get_task("task2") == task2
        assert dag.tasks == [task1, task2]


def test_add_task_list_to_predecessors(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task1.add_predecessors([task2, task3])
        assert task1.predecessors == set([task2, task3])
        assert dag.get_task("task2") == task2
        assert dag.get_task("task3") == task3
        assert dag.tasks == [task1, task2, task3]


def test_add_task_list_to_predecessors_ignoring_duplicates(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task1.add_predecessors([task2, task3])
        # Try to add the same list of tasks again; these duplicates will be ignored.
        task1.add_predecessors([task2, task3])
        assert task1.predecessors == set([task2, task3])
        assert dag.get_task("task2") == task2
        assert dag.get_task("task3") == task3
        assert dag.tasks == [task1, task2, task3]


def test_add_invalid_object_to_predecessors(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task = DAGTask("task1", "select 1")
        # DAGs themselves are not iterable.
        with pytest.raises(TypeError):
            task.add_predecessors(dag)


def test_add_one_successor(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task1.add_successors(task2)
        # Successors are not directly represented, but here task1 is a predecessor of task2.
        assert task2.predecessors == set([task1])
        assert dag.tasks == [task1, task2]


def test_add_task_list_to_successors(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task1 = DAGTask("task1", "select 1")
        task2 = DAGTask("task2", "select 2")
        task3 = DAGTask("task3", "select 3")
        task1.add_successors([task2, task3])
        # As above, successors are not directly represented, but task1 is now a predecessor of both task2 and
        # task3.
        assert task2.predecessors == set([task1])
        assert task3.predecessors == set([task1])
        assert dag.tasks == [task1, task2, task3]


def test_add_invalid_object_to_successors(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task = DAGTask("task1", "select 1")
        # DAGs themselves are not iterable.
        with pytest.raises(TypeError):
            task.add_successors(dag)


def test_illegal_move_task_to_different_dag(schema):
    dag1_name = random_object_name()
    dag2_name = random_object_name()
    with ExitStack() as dags:
        dags.enter_context(DAG(dag1_name, schedule=timedelta(minutes=10)))
        # dag1 is at the top of the DAG context stack so task1 automatically gets added to it.
        task1 = DAGTask("task1", "select 1")
        # Create dag2 and try to move task1 to it, which should fail.
        dag2 = dags.enter_context(DAG(dag2_name, schedule=timedelta(minutes=10)))
        with pytest.raises(ValueError):
            dag2.add_task(task1)


def test_dag_repr(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        assert repr(dag) == f"<DAG: {dag_name}>"


def test_dagtask_repr(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)):
        task = DAGTask("task1", "select 1")
        assert repr(task) == "<DAGTask: task1>"


def test_dag_contains_by_dagtask(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task = DAGTask("task1", "select 1")
        assert task in dag


def test_dag_contains_by_dag_name(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task = DAGTask("task1", "select 1")  # noqa: F841
        assert "task1" in dag


def test_dag_contains_by_dag_name_finalizer(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(minutes=10)) as dag:
        task = DAGTask("task1", "select 1", is_finalizer=True)  # noqa: F841
        assert "task1" in dag


def test_create_task_in_dag_without_warehouse(schema, warehouses, db_parameters):
    dag_name = random_object_name()
    dag = DAG(
        dag_name,
        schedule=timedelta(minutes=10),
        stage_location="fake_stage",
        warehouse=warehouses[db_parameters["warehouse"]].name,
    )
    with dag:
        task = DAGTask("task", "select 'task'")
    op = DAGOperation(schema)
    try:
        op.deploy(dag, mode=CreateMode.or_replace)
        op.run(dag)
        assert task.warehouse == warehouses[db_parameters["warehouse"]].name
    finally:
        op.drop(dag)


def test_create_task_in_dag_with_warehouse(schema, warehouses, db_parameters):
    dag_name = random_object_name()
    dag = DAG(
        dag_name,
        schedule=timedelta(minutes=10),
        stage_location="fake_stage",
        warehouse=warehouses[db_parameters["warehouse"]].name,
    )
    with dag:
        task = DAGTask("task", "select 'task'", warehouse=warehouses[db_parameters["warehouse"]].name)
    op = DAGOperation(schema)
    try:
        op.deploy(dag, mode=CreateMode.or_replace)
        op.run(dag)
        assert task.warehouse == warehouses[db_parameters["warehouse"]].name
    finally:
        op.drop(dag)


@pytest.mark.snowpark
def test_create_task_in_dag_with_different_warehouse(session, schema, warehouses, db_parameters):
    current_warehouse = session.get_current_warehouse()
    task_warehouse_name = f"SUB_TASK_WAREHOUSE_{random_object_name()}"
    warehouse = Warehouse(name=task_warehouse_name, warehouse_size="SMALL")
    warehouse_ref = None
    try:
        warehouse_ref = warehouses.create(warehouse, mode=CreateMode.or_replace)
        dag_name = random_object_name()
        dag = DAG(
            dag_name,
            schedule=timedelta(minutes=10),
            stage_location="fake_stage",
            warehouse=warehouses[db_parameters["warehouse"]].name,
        )
        with dag:
            task = DAGTask("task", "select 'task'", warehouse=task_warehouse_name)
        op = DAGOperation(schema)
        try:
            op.deploy(dag, mode=CreateMode.or_replace)
            op.run(dag)
            assert task.warehouse == task_warehouse_name
        finally:
            op.drop(dag)
    finally:
        warehouse_ref.drop()
        session.sql(f"USE WAREHOUSE {current_warehouse}").collect()


def test_create_serverless_task_in_dag_with_default_warehouse(schema, warehouses, db_parameters):
    dag_name = random_object_name()
    dag = DAG(
        dag_name,
        schedule=timedelta(minutes=10),
        stage_location="fake_stage",
        warehouse=warehouses[db_parameters["warehouse"]].name,
    )
    with dag:
        task = DAGTask("task", "select 'task'", is_serverless=True)
    op = DAGOperation(schema)
    try:
        op.deploy(dag, mode=CreateMode.or_replace)
        op.run(dag)
        assert task.warehouse is None
    finally:
        op.drop(dag)


def test_create_serverless_task_with_warehouse(schema, warehouses, db_parameters):
    dag_name = random_object_name()
    dag = DAG(
        dag_name,
        schedule=timedelta(minutes=10),
        stage_location="fake_stage",
        warehouse=warehouses[db_parameters["warehouse"]].name,
    )
    with pytest.raises(ValueError):
        with dag:
            _ = DAGTask(
                "task", "select 'task'", warehouse=warehouses[db_parameters["warehouse"]].name, is_serverless=True
            )


def test_create_task_without_warehouse_in_dag_without_warehouse(schema, warehouses, db_parameters):
    dag_name = random_object_name()
    dag = DAG(dag_name, schedule=timedelta(minutes=10), stage_location="fake_stage")
    with dag:
        task = DAGTask("task", "select 'task'", is_serverless=False)
    op = DAGOperation(schema)
    try:
        op.deploy(dag, mode=CreateMode.or_replace)
        op.run(dag)
        assert (
            task.warehouse is None
        )  # the task is created serverless anyway, as the DAG does not have warehouse either
    finally:
        op.drop(dag)


def test_create_dag_with_config(schema, database):
    dag_name = random_object_name()
    config = {"DB_NAME": database.name, "SCHEMA_NAME": schema.name}
    dag = DAG(dag_name, schedule=timedelta(days=1), config=config)
    with dag:
        op = DAGOperation(schema)
        try:
            op.deploy(dag, mode="orReplace")
            op.run(dag)
            dag_task_ref = schema.tasks[dag_name]
            dag_info = dag_task_ref.fetch()
            assert dag_info.config == config
        finally:
            op.drop(dag)


def test_create_dag_with_task_auto_retry_attempts(schema):
    dag_name = random_object_name()
    with DAG(dag_name, schedule=timedelta(days=1), task_auto_retry_attempts=1) as dag:
        op = DAGOperation(schema)
        try:
            op.deploy(dag, mode="orReplace")
            op.run(dag)
            dag_task_ref = schema.tasks[dag_name]
            dag_info = dag_task_ref.fetch()
            assert dag_info.task_auto_retry_attempts == 1
        finally:
            op.drop(dag)


# warnings are suppressed when we run tests in notebooks and stored procs
@pytest.mark.skip_notebook
@pytest.mark.skip_storedproc
def test_dag_with_user_task_managed_initial_warehouse_size_raises_deprecation_warning(schema):
    dag_name = random_object_name()
    with warnings.catch_warnings(record=True) as w:
        _ = DAG(dag_name, schedule=timedelta(days=1), user_task_managed_initial_warehouse_size="SMALL")
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "Providing user_task_managed_initial_warehouse_size for a dummy root Task is deprecated" in str(
            w[-1].message
        )


@pytest.mark.skip_notebook
@pytest.mark.skip_storedproc
def test_child_task_with_error_integration_raises_deprecation_warning(schema):
    dag_name = random_object_name()
    task_name = random_object_name()
    dag = DAG(dag_name, schedule=timedelta(days=1))
    with warnings.catch_warnings(record=True) as w:
        _ = DAGTask(task_name, "select 1", error_integration="non_existing_object", dag=dag)
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "error_integration cannot be specified for a child Task." in str(w[-1].message)


@pytest.mark.skip_notebook
@pytest.mark.skip_storedproc
def test_drop_with_drop_finalizer_as_false_raises_deprecation_warning(schema):
    with DAG("dag") as dag:
        with warnings.catch_warnings(record=True) as w:
            DAGTask("task", "select 1")
            op = DAGOperation(schema)
            op.deploy(dag)
            op.drop(dag, drop_finalizer=False)
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "Setting drop_finalizer to False is deprecated. " in str(w[-1].message)


def test_deploy_dag_with_to_sql(schema) -> None:
    class MLJobDefinition:
        def to_sql(self) -> str:
            return "CALL SYSTEM$EXECUTE_ML_JOB('test_ml_job')"

        def __call__(self) -> None:
            return None

    class OtherDefinition:
        def to_sql(self) -> str:
            return "select 1"

    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        ml_definition = MLJobDefinition()
        task_ml = DAGTask("task_ml", definition=ml_definition)
        other_definition = OtherDefinition()
        task_other = DAGTask("task_other", definition=other_definition)
        task_other >> task_ml
    low_level_task_ml = task_ml._to_low_level_task()
    assert low_level_task_ml.definition == "CALL SYSTEM$EXECUTE_ML_JOB('test_ml_job')"
    low_level_task_other = task_other._to_low_level_task()
    assert low_level_task_other.definition == "select 1"

    op = DAGOperation(schema)
    op.deploy(dag, mode=CreateMode.or_replace)
    try:
        # Fetch the deployed task and verify the definition was converted to SQL string
        fetched = schema.tasks[task_ml.full_name].fetch()
        assert fetched.definition == "CALL SYSTEM$EXECUTE_ML_JOB('test_ml_job')"

        # Verify the DAG structure was deployed correctly
        dependents = schema.tasks[test_dag].fetch_task_dependents()
        assert len(dependents) == 3
    finally:
        op.drop(dag)


def test_deploy_dag_with_to_sql_other_definition(schema) -> None:
    class OtherDefinition:
        def to_sql(self) -> None:
            return None

    test_dag = random_object_name()
    with DAG(test_dag, schedule=timedelta(minutes=10)) as dag:
        other_definition = OtherDefinition()
        _ = DAGTask("task_ml", definition=other_definition)
    op = DAGOperation(schema)
    with pytest.raises(TypeError):
        op.deploy(dag, mode=CreateMode.or_replace)



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from contextlib import suppress
from datetime import timedelta

import pytest

from pydantic_core._pydantic_core import ValidationError

from snowflake.core import CreateMode
from snowflake.core.exceptions import APIError, ConflictError, NotFoundError

from ..utils import random_object_name


task_name1 = random_object_name()
task_name2 = random_object_name()
task_name3 = random_object_name()

pytestmark = pytest.mark.snowpark


def test_get_task(tasks):
    with pytest.raises(NotFoundError):
        tasks["abc"].fetch()


def test_create_task_session_parameter(tasks):
    from snowflake.core.task import Task

    query = "select current_version()"
    try:
        # single parameter
        parameters = {"SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS": 80000}
        task = tasks.create(Task(name=task_name1, definition=query, session_parameters=parameters))
        fetched_task = task.fetch()
        assert fetched_task.session_parameters["SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS"] == 80000
        # multiple parameters
        parameters = {
            "SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS": 80000,
            "SNOWPARK_LAZY_ANALYSIS": False,
            "TIMEZONE": "America/New_York",
        }
        task = tasks.create(
            Task(name=task_name1, definition=query, session_parameters=parameters), mode=CreateMode.or_replace
        )
        parameters = task.fetch().session_parameters
        assert parameters["SNOWPARK_REQUEST_TIMEOUT_IN_SECONDS"] == 80000
        assert parameters["SNOWPARK_LAZY_ANALYSIS"] is False
        assert parameters["TIMEZONE"] == "America/New_York"

        # unsupported value
        with pytest.raises(TypeError) as ex_info:
            tasks.create(Task(name=task_name1, definition=query, session_parameters={"param": task}))
        ex_info.match(
            "Task.session_parameters is a dict. The value of this dict must be one of str, int, float, or bool."
        )
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_overwrite(tasks):
    from snowflake.core.task import Task

    try:
        query = "select current_version()"
        task_info = tasks.create(Task(name=task_name1, definition=query)).fetch()
        assert task_name1 == task_info.name.lower()
        assert query in task_info.definition

        with pytest.raises(ConflictError) as ex_info:
            tasks.create(Task(name=task_name1, definition=query))
        assert ex_info.value.status == 409

        created_time = task_info.created_on
        # no error
        task = tasks.create(Task(name=task_name1, definition=query), mode="orreplace")
        # should be the new one
        new_info = task.fetch()
        assert created_time != new_info.created_on
        created_time = new_info.created_on
        tasks.create(Task(name=task_name1, definition=query), mode="ifnotexists")
        # should be the old one
        assert created_time == task.fetch().created_on

    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_warehouse(tasks, db_parameters):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", warehouse=db_parameters["warehouse"])
        )
        assert task.fetch().warehouse.lower() in db_parameters["warehouse"].lower()
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_warehouse_size(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(
                name=task_name1, definition="select current_version()", user_task_managed_initial_warehouse_size="LARGE"
            )
        )
        assert task.fetch().user_task_managed_initial_warehouse_size == "LARGE"
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


@pytest.mark.min_sf_ver("9.4.0")
def test_create_task_target_completion_interval(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", target_completion_interval=timedelta(hours=2))
        )
        assert task.fetch().target_completion_interval == timedelta(minutes=120)
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


@pytest.mark.min_sf_ver("9.4.0")
def test_create_task_serverless_task_min_statement_size(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", serverless_task_min_statement_size="MEDIUM")
        )
        assert task.fetch().serverless_task_min_statement_size == "MEDIUM"
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


@pytest.mark.min_sf_ver("9.4.0")
def test_create_task_serverless_task_max_statement_size(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", serverless_task_max_statement_size="XLARGE")
        )
        assert task.fetch().serverless_task_max_statement_size == "XLARGE"
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_schedule_cron(tasks):
    from snowflake.core.task import Cron, Task

    try:
        schedule = Cron("0 9-17 * * SUN", "America/Los_Angeles")
        task = tasks.create(Task(name=task_name1, definition="select current_version()", schedule=schedule))
        assert task.fetch().schedule == schedule
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_schedule_timedelta(tasks):
    from snowflake.core.task import Task

    try:
        with pytest.raises(ValueError) as ex_info:
            tasks.create(
                Task(name=task_name1, definition="select current_version()", schedule=timedelta(microseconds=1))
            )
        assert ex_info.match("The schedule time delta must be")

        with pytest.raises(ValueError) as ex_info:
            tasks.create(Task(name=task_name1, definition="select current_version()", schedule=timedelta(weeks=2)))
        assert ex_info.match("The schedule time delta must be")

        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", schedule=timedelta(minutes=11))
        )
        assert task.fetch().schedule.total_seconds() == 660

    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_allow_overlapping_execution(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", allow_overlapping_execution=True)
        ).fetch()
        assert task.allow_overlapping_execution
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


@pytest.mark.parametrize("comment", ["test_comment", "test comment", "TEST*COMMENT", None])
def test_create_task_comment(tasks, comment):
    from snowflake.core.task import Task

    try:
        task = tasks.create(Task(name=task_name1, definition="select current_version()", comment=comment)).fetch()
        assert task.comment == (comment if comment else "")
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_user_task_timeout_ms(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(Task(name=task_name1, definition="select current_version()", user_task_timeout_ms=1234))
        assert task.fetch().user_task_timeout_ms == 1234
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_suspend_task_after_num_failures(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", suspend_task_after_num_failures=1234)
        )
        assert task.fetch().suspend_task_after_num_failures == 1234
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


def test_create_task_predecessors(tasks):
    from snowflake.core.task import Task
    from snowflake.snowpark._internal.utils import parse_table_name

    try:
        task_name1_with_special_char = '"a b"'
        task1 = tasks.create(Task(name=task_name1_with_special_char, definition="select current_version()"))

        task2 = tasks.create(Task(name=task_name2, definition="select current_version()"))

        task3 = tasks.create(
            Task(name=task_name3, definition="select current_version()", predecessors=[task1.name, task2.name])
        )

        task3_predecessors = sorted(map(lambda fqn: parse_table_name(fqn)[-1], task3.fetch().predecessors))
        expected = sorted([tasks[task_name2].fetch().name, tasks[task_name1_with_special_char].fetch().name])
        assert expected == task3_predecessors
    finally:
        with suppress(Exception):
            tasks[task_name1_with_special_char].drop()
        with suppress(Exception):
            tasks[task_name2].drop()
        with suppress(Exception):
            tasks[task_name3].drop()


def test_create_task_condition(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(
                name=task_name1, definition="select current_version()", condition="SYSTEM$STREAM_HAS_DATA('my_stream')"
            )
        ).fetch()
        assert task.condition == "SYSTEM$STREAM_HAS_DATA('my_stream')"
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


@pytest.mark.usefixtures("my_integration_exists")
def test_create_task_error_integration(tasks):
    from snowflake.core.task import Task

    try:
        task = tasks.create(
            Task(name=task_name1, definition="select current_version()", error_integration="my_integration")
        ).fetch()
        assert task.error_integration == "my_integration"
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()


# not enabled on jenkins for now, since task tests need some extra setup which is under construction
def test_create_task_task_auto_retry_attempts(tasks):
    from snowflake.core.task import Task

    test_cases = [None, 0, 1, -1, 31]
    try:
        for val in test_cases:
            params = {"name": task_name1, "definition": "select current_version()"}
            if val is not None:
                params["task_auto_retry_attempts"] = val
            try:
                task = tasks.create(Task(**params), mode=CreateMode.or_replace).fetch()
                if val is None:
                    assert task.task_auto_retry_attempts is None
                elif 0 <= val <= 30:
                    assert task.task_auto_retry_attempts == val
                else:
                    pytest.fail("task creation should fail. Valid value range is 0 - 30")
            except ValidationError:
                assert val < 0 or val > 30

        # test child task
        with pytest.raises(APIError) as error:
            tasks.create(
                Task(
                    name=task_name2,
                    definition="select current_version()",
                    predecessors=[task_name1],
                    task_auto_retry_attempts=1,
                )
            )
            assert error.status == 400
            assert "cannot set parameter task_auto_retry_attempts on non-root task" in error.body
    finally:
        with suppress(NotFoundError):
            tasks[task_name1].drop()



"""Manages Snowflake Notebooks.

Examples:
    >>> notebooks: NotebookCollection = root.databases["my_db"].schemas["my_schema"].notebooks
    >>> my_notebook = notebooks.create(Notebook("my_notebook"))
    >>> notebook_iter = notebooks.iter(like="my%")
    >>> notebook = notebooks["my_notebook"]
    >>> an_existing_notebook = notebooks["an_existing_notebook"]

Refer to :class:`snowflake.core.Root` to create the ``root``.
"""

from ._notebook import Notebook, NotebookCollection, NotebookResource, VersionDetails


__all__ = ["Notebook", "NotebookCollection", "NotebookResource", "VersionDetails"]



import pytest

from snowflake.core.exceptions import NotFoundError


def test_resume_and_suspend_cluster(tables, table_handle):
    # verify resume
    table_handle.resume_recluster()
    table_handle.drop()
    table_handle.resume_recluster(if_exists=True)
    with pytest.raises(NotFoundError):
        table_handle.resume_recluster()
    table_handle.undrop()
    # check resume after suspend
    table_handle.suspend_recluster()
    table_handle.resume_recluster()

    # verify suspend
    # verify suspend when it is resumed
    table_handle.suspend_recluster()

    # verify suspend when it is suspended
    table_handle.suspend_recluster()

    # verify suspend when it is dropped
    table_handle.drop()
    with pytest.raises(NotFoundError):
        table_handle.suspend_recluster()
    table_handle.suspend_recluster(if_exists=True)



from contextlib import suppress

import pytest as pytest

from snowflake.core.database import Database
from snowflake.core.database_role import DatabaseRole
from snowflake.core.exceptions import NotFoundError
from snowflake.core.role import Role
from snowflake.core.schema import Schema
from snowflake.core.table import Table, TableColumn
from snowflake.core.user import User
from tests.integ.utils import random_string


@pytest.fixture
def test_database_role_name(database_roles_in_test_database):
    role_name = random_string(4, "test_database_role_name_")
    test_role = DatabaseRole(name=role_name, comment="test_comment")
    try:
        database_roles_in_test_database.create(test_role)
        yield role_name.upper()
    finally:
        with suppress(NotFoundError):
            database_roles_in_test_database[role_name].drop()


@pytest.fixture
def test_database_role_name_2(database_roles_in_test_database):
    role_name = random_string(4, "test_database_role_name_2_")
    test_role = DatabaseRole(name=role_name, comment="test_comment")
    try:
        database_roles_in_test_database.create(test_role)
        yield role_name.upper()
    finally:
        with suppress(NotFoundError):
            database_roles_in_test_database[role_name].drop()


@pytest.fixture
def test_database_for_grant_name(databases, backup_database_schema):
    del backup_database_schema
    try:
        new_db_def = Database(name=random_string(3, "test_database_for_grant_name_"), kind="TRANSIENT")
        new_db_def.comment = "database first"
        database = databases.create(new_db_def)
        yield database.name.upper()
    finally:
        with suppress(Exception):
            database.drop()


@pytest.fixture
def database_roles_in_test_database(databases, test_database_for_grant_name):
    return databases[test_database_for_grant_name].database_roles


@pytest.fixture
def schemas_in_test_database(databases, test_database_for_grant_name):
    return databases[test_database_for_grant_name].schemas


@pytest.fixture
def tables_in_test_database(schemas_in_test_database, test_schema_for_grant_name):
    return schemas_in_test_database[test_schema_for_grant_name].tables


@pytest.fixture
def test_schema_for_grant_name(schemas_in_test_database):
    schema_name = random_string(5, "test_schema_for_role_grant_name_")
    try:
        new_schema_def = Schema(name=schema_name)
        schema = schemas_in_test_database.create(new_schema_def)
        yield schema_name.upper()
    finally:
        with suppress(NotFoundError):
            schema.drop()


@pytest.fixture
def test_table_for_grant_name(tables_in_test_database):
    table_name = random_string(5, "test_table_for_test_role_")
    columns = [TableColumn(name="col1", datatype="int"), TableColumn(name="col2", datatype="string")]
    test_table = Table(name=table_name, columns=columns)
    try:
        test_table_handle = tables_in_test_database.create(test_table)
        yield table_name.upper()
    finally:
        with suppress(NotFoundError):
            test_table_handle.drop()


@pytest.fixture
def test_role_name(roles):
    role_name = random_string(4, "test_grant_role_")
    test_role = Role(name=role_name, comment="test_comment")
    try:
        roles.create(test_role)
        yield role_name
    finally:
        with suppress(Exception):
            roles[role_name].drop()


@pytest.fixture
def test_role_name_2(roles):
    role_name = random_string(4, "test_grant_role_2_")
    test_role = Role(name=role_name, comment="test_comment")
    try:
        roles.create(test_role)
        yield role_name
    finally:
        with suppress(Exception):
            roles[role_name].drop()


@pytest.fixture
def test_user_name(users):
    user_name = random_string(4, "test_grant_user_")
    test_user = User(
        name=user_name,
        password="test",
        display_name="test_name",
        first_name="firstname",
        last_name="lastname",
        email="test@snowflake.com",
        must_change_password=False,
        disabled=False,
        days_to_expiry=1,
        mins_to_unlock=10,
        mins_to_bypass_mfa=60,
        default_warehouse="test",
        default_namespace="test",
        default_role="public",
        default_secondary_roles="ALL",
        comment="test_comment",
    )
    try:
        users.create(test_user)
        yield user_name
    finally:
        with suppress(Exception):
            users[user_name].drop()



import pytest

from tests.integ.dynamic_table.util import assert_dynamic_table


@pytest.mark.min_sf_ver("8.27.0")
def test_fetch(dynamic_table_handle, table_handle, database, schema, db_parameters):
    dynamic_table = dynamic_table_handle.fetch()
    assert_dynamic_table(dynamic_table, dynamic_table_handle.name, table_handle, database, schema, db_parameters, True)



import json
import logging
import os
import warnings

from json import JSONDecodeError
from typing import TYPE_CHECKING, Optional, Union

from snowflake.connector import Error, SnowflakeConnection, cursor
from snowflake.core._options import require_snowpark

from ._common import TokenType
from ._constants import PLATFORM, PYTHON_VERSION
from ._internal.client_info import ClientInfo
from ._internal.root_configuration import RootConfiguration
from ._internal.snowapi_parameters import SnowApiParameter, SnowApiParameters
from ._internal.telemetry import ApiTelemetryClient
from ._utils import fix_hostname as _fix_hostname
from .account import AccountCollection
from .api_integration import ApiIntegrationCollection
from .catalog_integration import CatalogIntegrationCollection
from .compute_pool import ComputePoolCollection
from .cortex.chat_service import CortexChatService
from .cortex.embed_service import CortexEmbedService
from .cortex.inference_service import CortexInferenceService
from .cortex.lite_agent_service import CortexAgentService
from .database import DatabaseCollection
from .external_volume import ExternalVolumeCollection
from .grant._grants import Grants
from .managed_account import ManagedAccountCollection
from .network_policy import NetworkPolicyCollection
from .notification_integration import NotificationIntegrationCollection
from .role import RoleCollection
from .session import SnowAPISession
from .user import UserCollection
from .version import __version__
from .warehouse import WarehouseCollection


if TYPE_CHECKING:
    from snowflake.snowpark import Session

logger = logging.getLogger(__name__)


class Root:
    """The entry point of the Snowflake Core Python APIs that manage the Snowflake objects.

    Parameters
    __________
    connection: Union[SnowflakeConnection, Session]
        A ``SnowflakeConnection`` or Snowpark ``Session`` instance.

    Examples
    ________
    Creating a ``Root`` instance:

    >>> from snowflake.connector import connect
    >>> from snowflake.core import Root
    >>> from snowflake.snowpark import Session
    >>> CONNECTION_PARAMETERS = {
    ...     "account": os.environ["snowflake_account_demo"],
    ...     "user": os.environ["snowflake_user_demo"],
    ...     "password": os.environ["snowflake_password_demo"],
    ...     "database": test_database,
    ...     "warehouse": test_warehouse,
    ...     "schema": test_schema,
    ... }
    >>> # create from a Snowflake Connection
    >>> connection = connect(**CONNECTION_PARAMETERS)
    >>> root = Root(connection)
    >>> # or create from a Snowpark Session
    >>> session = Session.builder.config(CONNECTION_PARAMETERS).create()
    >>> root = Root(session)

    Using the ``Root`` instance to access resource management APIs:

    >>> tasks = root.databases["mydb"].schemas["myschema"].tasks
    >>> mytask = tasks["mytask"]
    >>> mytask.resume()
    >>> compute_pools = root.compute_pools
    >>> my_computepool = compute_pools["mycomputepool"]
    >>> my_computepool.delete()
    """

    def __init__(
        self, connection: Union[SnowflakeConnection, "Session"], root_config: Optional[RootConfiguration] = None
    ) -> None:
        self._session: Optional[Session] = None
        if isinstance(connection, SnowflakeConnection):
            self._connection = connection
        else:
            self._session = connection
            self._connection = connection._conn._conn

        logger.info("New root object was created for %r", connection)

        self._refresh_parameters()
        # In stored procedures connection.rest doesn't have _host. Let's set the base-case here.
        self._hostname = None
        if self._connection.rest is not None and hasattr(self._connection.rest, "_host"):
            self._hostname = self._connection.rest._host
            if self._parameters.fix_hostname:
                # Fix hostname, if desired. This is done by replacing _ (underscores) with - (dashes)
                #  in the first part of the hostname.
                self._hostname = _fix_hostname(self._hostname)
        self._snowapi_session = SnowAPISession(self)
        self._initialize_client_info()

        self._databases = DatabaseCollection(self)
        self._accounts = AccountCollection(self)
        self._managed_accounts = ManagedAccountCollection(self)
        self._compute_pools = ComputePoolCollection(self)
        self._external_volumes = ExternalVolumeCollection(self)
        self._telemetry_client = ApiTelemetryClient(self._connection)
        self._warehouses = WarehouseCollection(self)
        self._network_policies = NetworkPolicyCollection(self)
        self._roles = RoleCollection(self)
        self._grants = Grants(self)
        self._users = UserCollection(self)
        self._catalog_integrations = CatalogIntegrationCollection(self)
        self._notification_integrations = NotificationIntegrationCollection(self)
        self._api_integrations = ApiIntegrationCollection(self)

        self._cortex_chat_service = CortexChatService(self)
        self._cortex_inference_service = CortexInferenceService(self)
        self._cortex_embed_service = CortexEmbedService(self)
        self._cortex_agent_service = CortexAgentService(self)

        logger.info("Snowflake Core version: %s, on Python %s, on platform: %s", __version__, PYTHON_VERSION, PLATFORM)
        parameter_map = self._parameters.params_map
        for parameter in parameter_map:
            logger.info("Parameter %s: %s", parameter, parameter_map.get(parameter))

        self._root_config = root_config or RootConfiguration()

    def parameters(self, refresh: bool = False) -> SnowApiParameters:
        if refresh:
            self._refresh_parameters()

        return self._parameters

    @property
    def root_config(self) -> RootConfiguration:
        """Return the root configuration object."""
        return self._root_config

    @property
    def connection(self) -> SnowflakeConnection:
        """Return the connection in use.

        This is the connection used to create this ``Root`` instance, or the
        Snowpark session's connection if this root is created from a
        session.
        """
        return self._connection

    @property
    def session(self) -> "Session":
        """Returns the session that is used to create this ``Root`` instance."""
        require_snowpark()
        if self._session is None:
            from snowflake.snowpark.session import Session, _active_sessions

            self._session = Session.builder.configs({"connection": self._connection}).create()
            _active_sessions.remove(self._session)  # This is supposed to avoid a user double using sessions
        return self._session

    @property
    def databases(self) -> DatabaseCollection:
        """Returns the ``DatabaseCollection`` that represents the visible databases.

        Examples
        ________

        Getting a specific database resource:

        >>> root = Root(session)
        >>> my_db = root.databases["my_db"]
        """
        return self._databases

    @property
    def accounts(self) -> AccountCollection:
        """Returns the ``AccountCollection`` that represents the visible accounts.

        Examples
        ________

        Getting a specific account resource:

        >>> root = Root(session)
        >>> my_account = root.accounts["my_account"]
        """
        return self._accounts

    @property
    def api_integrations(self) -> ApiIntegrationCollection:
        """Returns the ``ApiIntegrationCollection`` that represents the visible API integrations.

        Examples
        ________

        Getting a specific API integration resource:

        >>> root = Root(session)
        >>> my_api_int = root.api_integrations["my_api_int"]
        """
        return self._api_integrations

    @property
    def cortex_inference_service(self) -> CortexInferenceService:
        """Returns the ``CortexInferenceService`` that represents the cortex inference service.

        Examples
        ________

        Getting the cortex inference service resource:

        >>> root = Root(session)
        >>> my_cortex_inference_service = root.cortex_inference_service
        """
        return self._cortex_inference_service

    @property
    def cortex_chat_service(self) -> CortexChatService:
        """
        Returns the CortexChatService that represents the cortex chat service.

        Examples
        ________
        To get the cortex chat service resource, you can do the following:

        >>> root = Root(session)
        >>> my_cortex_chat_service = root.cortex_chat_service
        """
        return self._cortex_chat_service

    @property
    def cortex_embed_service(self) -> CortexEmbedService:
        """
        Returns the CortexEmbedService that represents the cortex embed service.

        Examples
        ________
        To get the cortex embed service resource, you can do the following:

        >>> root = Root(session)
        >>> my_cortex_embed_service = root.cortex_embed_service
        """
        return self._cortex_embed_service

    @property
    def cortex_agent_service(self) -> CortexAgentService:
        """
        Returns the CortexAgentService that represents the cortex lite Agent service.

        Examples
        ________
        To get the cortex lite Agent service resource, you can do the following:

        >>> root = Root(session)
        >>> my_cortex_agent_service = root.cortex_agent_service
        """
        return self._cortex_agent_service

    @property
    def managed_accounts(self) -> ManagedAccountCollection:
        """Returns the ``ManagedAccountCollection`` that represents the visible accounts.

        Examples
        ________

        Getting a specific managed account resource:

        >>> root = Root(session)
        >>> my_managed_account = root.managed_accounts["my_managed_account"]
        """
        return self._managed_accounts

    @property
    def compute_pools(self) -> ComputePoolCollection:
        """Returns the ``ComputePoolCollection`` that represents the visible compute pools.

        Examples
        ________

        Getting a specific compute pool resource:

        >>> root = Root(session)
        >>> my_cp = root.compute_pools["my_cp"]
        """
        return self._compute_pools

    @property
    def external_volumes(self) -> ExternalVolumeCollection:
        """Returns the ``ExternalVolumeCollection`` that represents the visible external volumes.

        Examples
        ________

        Getting a specific external volume resource:

        >>> root = Root(session)
        >>> my_external_volume = root.external_volumes["my_external_volume"]
        """
        return self._external_volumes

    @property
    def network_policies(self) -> NetworkPolicyCollection:
        """Returns the ``NetworkPolicyCollection`` that represents the visible network policies.

        Examples
        ________

        Getting a specific network policy resource:

        >>> root = Root(session)
        >>> my_network_policy = root.network_policies["my_network_policy"]
        """
        return self._network_policies

    @property
    def notification_integrations(self) -> NotificationIntegrationCollection:
        """Returns the ``NotificationIntegrationCollection`` that represents the visible notification integrations.

        Examples
        ________

        Listing all available Notification Integrations:

        >>> root = Root(session)
        >>> my_nis = list(root.notification_integrations.iter())
        """
        return self._notification_integrations

    @property
    def warehouses(self) -> WarehouseCollection:
        """Returns the ``WarehouseCollection`` that represents the visible warehouses.

        Examples
        ________

        Getting a specific warehouse resource:

        >>> root = Root(session)
        >>> my_wh = root.warehouses["my_wh"]
        """
        return self._warehouses

    @property
    def roles(self) -> RoleCollection:
        """Returns the ``RoleCollection`` that represents the visible roles.

        Examples
        ________

        Getting a specific role resource:

        >>> root = Root(session)
        >>> my_role = root.roles["my_role"]
        """
        return self._roles

    @property
    def grants(self) -> Grants:
        """Returns the visible Grants in Snowflake.

        Examples
        ________

        Using the ``Grants`` object to grant a privilege to a role:

        >>> grants.grant(
        ...     Grant(
        ...         grantee=Grantees.role(name="public"),
        ...         securable=Securables.database("invaliddb123"),
        ...         privileges=[Privileges.create_database],
        ...         grant_option=False,
        ...     )
        ... )
        """
        return self._grants

    @property
    def users(self) -> UserCollection:
        """Returns the ``UserCollection`` that represents the visible users.

        Examples
        ________

        Getting a specific user resource:

        >>> root = Root(session)
        >>> my_user = root.users["my_user"]
        """
        return self._users

    @property
    def catalog_integrations(self) -> CatalogIntegrationCollection:
        """Returns the ``CatalogIntegrationCollection`` that represents the visible catalog integrations.

        Examples
        ________

        Getting a specific catalog integration resource:

        >>> root = Root(session)
        >>> my_catalog_integration = root.catalog_integrations["my_catalog_integration"]
        """
        return self._catalog_integrations

    @property
    def _session_token(self) -> Optional[str]:
        # TODO: this needs to be fixed in the connector
        return self._connection.rest.token  # type: ignore[union-attr]

    @property
    def _master_token(self) -> Optional[str]:
        # TODO: this needs to be fixed in the connector
        return self._connection.rest.master_token  # type: ignore[union-attr]

    @property
    def external_session_id(self) -> Optional[str]:
        return getattr(self._connection.rest, "_external_session_id", None)

    @property
    def token_type(self) -> TokenType:
        if getattr(self._connection, "_authenticator", None) == "PAT_WITH_EXTERNAL_SESSION":
            return TokenType.EXTERNAL_SESSION_WITH_PAT
        return TokenType.SESSION_TOKEN

    def _refresh_parameters(self) -> None:
        parameters = {
            SnowApiParameter.USE_CLIENT_RETRY: os.getenv("_SNOWFLAKE_ENABLE_RETRY_REQUEST_QUERY"),
            SnowApiParameter.PRINT_VERBOSE_STACK_TRACE: os.getenv("_SNOWFLAKE_PRINT_VERBOSE_STACK_TRACE"),
            SnowApiParameter.FIX_HOSTNAME: os.getenv("_SNOWFLAKE_FIX_HOSTNAME", "true"),
            SnowApiParameter.MAX_THREADS: os.getenv("_SNOWFLAKE_MAX_THREADS"),
        }
        self._parameters = SnowApiParameters(parameters)

    def _query_for_client_info(self) -> Optional[list[dict[str, str]]]:
        logger.info("Querying the server for client info")
        ret = None
        query_res = self.connection.cursor().execute("SELECT SYSTEM$CLIENT_VERSION_INFO()")
        if query_res is not None and type(query_res) is cursor.SnowflakeCursor:
            result = query_res.fetchone()
            if isinstance(result, tuple):
                ret = json.loads(result[0])

        return ret

    def _initialize_client_info(self) -> None:
        client_support_info = None

        try:
            client_info_from_query = self._query_for_client_info()

            if client_info_from_query is not None:
                # Find the entry that corresponds to PyCore. If not found, then we assume all versions
                # are supported; this will be None and ClientInfo will be instantiated with None
                # as the value for the support info.
                client_support_info = next(
                    (client for client in client_info_from_query if client.get("clientId") == "PyCore"), None
                )
        except (Error, JSONDecodeError) as error:
            warning = "Could not fetch client version info from Snowflake. Please make sure your client is up to date."
            logger.warning(warning, exc_info=error)
            warnings.warn(warning, stacklevel=3)

        client_info = ClientInfo(client_support_info)

        if not client_info.version_is_supported():
            # Later on, we may make this more aggressive and terminate with an error.
            warning = (
                f"Your current version of the client, {client_info.client_version}, "
                f"is not supported. The minimum supported version is {client_info.minimum_supported_version}."
            )

            if client_info.recommended_version:
                warning = warning + f" The recommended version is {client_info.recommended_version}."

            logger.warning(warning)
            warnings.warn(warning, stacklevel=3)

        elif client_info.version_is_nearing_end_of_support():
            warning = (
                f"Your current version of the client, {client_info.client_version}, is nearing the end of support."
            )

            if client_info.recommended_version:
                warning = warning + f" The recommended version is {client_info.recommended_version}."

            logger.warning(warning)
            warnings.warn(warning, stacklevel=3)

        self._client_info = client_info



"""
Snowflake Streamlit API.

The Snowflake Streamlit API is a REST API that you can use to access, update, and perform certain actions on Streamlit resources in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict

from snowflake.core.streamlit._generated.models.commit_streamlit_request_version import (
    CommitStreamlitRequestVersion,
    CommitStreamlitRequestVersionModel,
)


class CommitStreamlitRequest(BaseModel):
    """A model object representing the CommitStreamlitRequest resource.

    Constructs an object of type CommitStreamlitRequest with the provided properties.

    Parameters
    __________
    version : CommitStreamlitRequestVersion, optional
    """

    version: Optional[CommitStreamlitRequestVersion] = None

    __properties = ["version"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> CommitStreamlitRequest:
        """Create an instance of CommitStreamlitRequest from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # override the default output from pydantic by calling `to_dict()` of version
        if self.version:
            _dict["version"] = self.version.to_dict()

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> CommitStreamlitRequest:
        """Create an instance of CommitStreamlitRequest from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return CommitStreamlitRequest.model_validate(obj)

        _obj = CommitStreamlitRequest.model_validate(
            {
                "version": CommitStreamlitRequestVersion.from_dict(obj.get("version"))
                if obj.get("version") is not None
                else None,
            }
        )

        return _obj


class CommitStreamlitRequestModel:
    def __init__(
        self,  # optional properties
        version: Optional[CommitStreamlitRequestVersion] = None,
    ):
        """A model object representing the CommitStreamlitRequest resource.

        Constructs an object of type CommitStreamlitRequest with the provided properties.

        Parameters
        __________
        version : CommitStreamlitRequestVersion, optional
        """
        self.version = version

    __properties = ["version"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return CommitStreamlitRequest(
            version=self.version._to_model() if self.version is not None else None,
        )

    @classmethod
    def _from_model(cls, model) -> CommitStreamlitRequestModel:
        return CommitStreamlitRequestModel(
            version=CommitStreamlitRequestVersionModel._from_model(model.version) if model.version else None,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a CommitStreamlitRequest.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the CommitStreamlitRequest object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> CommitStreamlitRequestModel:
        """Create an instance of CommitStreamlitRequest from a dict.

        This method constructs a CommitStreamlitRequest object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        CommitStreamlitRequest
            A CommitStreamlitRequest object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(CommitStreamlitRequest.from_dict(obj))


CommitStreamlitRequest._model_class = CommitStreamlitRequestModel



from __future__ import annotations

import json
import logging
import re

from decimal import Decimal
from typing import TYPE_CHECKING, Any, Callable, Optional, Protocol, TypeVar

from pydantic import StrictStr

from snowflake.core.exceptions import InvalidResultError


if TYPE_CHECKING:
    from snowflake.core import Root
    from snowflake.core.function import Function
    from snowflake.core.procedure import Procedure
    from snowflake.core.tag import TagResource, TagValue
    from snowflake.core.user_defined_function import UserDefinedFunction

logger = logging.getLogger(__name__)


def get_function_name_with_args(function: Function | Procedure | UserDefinedFunction) -> str:
    return f"{function.name}({','.join([str(argument.datatype) for argument in function.arguments])})"


FUNCTION_WITH_ARGS_PATTERN = re.compile(r"""^(\"([^\"]|\"\")+\"|[a-zA-Z_][a-zA-Z0-9_$]*)\(([A-Za-z,0-9_]*)\)$""")


def replace_function_name_in_name_with_args(name_with_args: str, new_name: str) -> str:
    matcher = FUNCTION_WITH_ARGS_PATTERN.match(name_with_args)
    if not matcher:
        raise ValueError("Invalid function name with arguments")

    args = matcher.group(3)
    return f"{new_name}({args})"


def check_version_gte(version_to_check: str, reference_version: str) -> bool:
    cur_version = tuple(map(int, version_to_check.split(".")))
    req_version = tuple(map(int, reference_version.split(".")))

    return cur_version >= req_version


def check_version_lte(version_to_check: str, reference_version: str) -> bool:
    cur_version = tuple(map(int, version_to_check.split(".")))
    req_version = tuple(map(int, reference_version.split(".")))

    return cur_version <= req_version


def fix_hostname(hostname: str) -> str:
    """Perform automatic hostname fixes.

    When a legacy format hostname is used to connect to Snowflake SSL certificates might not work as expected.
    For example if the legacy url format is used to connect to Snowflake (see
    https://docs.snowflake.com/en/user-guide/organizations-connect for more documentation) _ (underscores) should
    be replaced with - (dashes).
    """
    first_part, _, rest = hostname.partition(".")
    if not rest:
        raise Exception(f"provided hostname '{hostname}' is invalid")
    first_part = first_part.replace("_", "-")
    new_hostname = f"{first_part}.{rest}"
    if hostname != new_hostname:
        logger.info(f"Hostname automatically fixed from '{hostname}' to '{new_hostname}'")
    return new_hostname


def cast_result(result: str, datatype: str | None) -> Any:
    if datatype in ["INT", "INTEGER", "BIGINT", "SMALLINT", "TINYINT", "BYTEINT"]:
        return int(result)
    if datatype in ["NUMBER", "DECIMAL", "NUMERIC"]:
        try:
            return int(result)
        except ValueError:
            return float(result)
    if datatype in ["FLOAT", "FLOAT4", "FLOAT8", "DOUBLE", "DOUBLE PRECISION", "REAL"]:
        return float(result)
    if datatype == "DECFLOAT":
        return Decimal(result)
    if datatype in ["VARCHAR", "STRING", "TEXT"]:
        return str(result)
    if datatype in ["CHAR", "CHARACTER"]:
        return result
    if datatype == "ARRAY":
        return list(result)
    if datatype in ["GEOMETRY", "GEOGRAPHY", "OBJECT", "VARIANT"]:
        return json.loads(result)
    if datatype == "BOOLEAN":
        return result.lower() in ("yes", "y", "true", "t", "1")
    return result


def map_result(procedure: Procedure, raw_result: Any, extract: bool = False) -> Any:
    """Map result from API to native Python types."""
    from snowflake.core.procedure import ReturnDataType, ReturnTable

    if not isinstance(raw_result, list):
        raise InvalidResultError(f"Procedure result {str(raw_result)} is invalid or empty")

    rt_type = procedure.return_type
    if isinstance(rt_type, ReturnTable):
        if rt_type.column_list is None:
            return raw_result
        processed_rows = []
        columns_mapping = {c.name.lower(): c.datatype for c in rt_type.column_list}
        for row in raw_result:
            processed_rows.append({k: cast_result(v, columns_mapping.get(k.lower())) for k, v in row.items()})
        return processed_rows

    if isinstance(rt_type, ReturnDataType):
        payload = raw_result[0]
        if not isinstance(payload, dict):
            raise TypeError(f"Expected first item to be of type dict but got {type(payload)}")

        if not extract:
            return [{k: cast_result(v, rt_type.datatype) for k, v in payload.items()}]

        # Unpack the result of [{sproc_name: RESULT}]
        result = payload[next(iter(payload.keys()))]
        return cast_result(result, rt_type.datatype)
    return raw_result


class TagAssignmentLike(Protocol):
    tag_value: StrictStr
    level: Optional[StrictStr]
    tag_database: Optional[str]
    tag_name: str
    tag_schema: Optional[str]


TTagAssignment = TypeVar("TTagAssignment", bound=TagAssignmentLike)
TTagReference = TypeVar("TTagReference")


def tag_tuple_to_tag_assignment(
    tag_assignment_cls: Callable[..., TTagAssignment],
    tag_resource: TagResource,
    tag_value: TagValue,
) -> TTagAssignment:
    return tag_assignment_cls(
        tag_name=tag_resource.name,
        tag_schema=tag_resource.schema.name,
        tag_database=tag_resource.database.name,
        tag_value=tag_value.value,
    )


def tag_resource_to_tag_reference(
    tag_reference_cls: Callable[..., TTagReference], resource: TagResource
) -> TTagReference:
    return tag_reference_cls(
        tag_name=resource.name,
        tag_schema=resource.schema.name,
        tag_database=resource.database.name,
    )


def tag_assignment_to_tag_tuple(ta: TagAssignmentLike, root: Root) -> tuple[TagResource, TagValue]:
    from snowflake.core.tag import TagValue

    db = ta.tag_database
    schema = ta.tag_schema

    if db is None:
        raise ValueError("TagAssignment must have tag_database set.")

    if schema is None:
        raise ValueError("TagAssignment must have tag_schema set.")

    tag_resource = root.databases[db].schemas[schema].tags[ta.tag_name]
    tag_value = TagValue(ta.tag_value, ta.level)
    return tag_resource, tag_value



"""
Snowflake Function API.

The Snowflake Function API is a REST API that allows caller to create, execute and drop functions in a Snowflake database.  # noqa: E501

The version of the OpenAPI document: 0.0.1
Contact: support@snowflake.com
Generated by: https://openapi-generator.tech

Do not edit this file manually.
"""

from __future__ import annotations

import json
import pprint
import re  # noqa: F401

from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, StrictStr


class FunctionArgument(BaseModel):
    """A model object representing the FunctionArgument resource.

    Constructs an object of type FunctionArgument with the provided properties.

    Parameters
    __________
    name : str, optional
        Argument's name
    datatype : str,  default 'TEXT'
        Argument's type
    value : object, optional
        Argument's value
    """

    name: Optional[StrictStr] = None

    datatype: Optional[StrictStr] = "TEXT"

    value: Optional[Any] = None

    __properties = ["name", "datatype", "value"]

    model_config = ConfigDict(
        validate_by_name=True,
        validate_assignment=True,
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> FunctionArgument:
        """Create an instance of FunctionArgument from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(
        self,
        hide_readonly_properties: bool = False,
    ) -> dict[str, Any]:
        """Returns the dictionary representation of the model using alias."""
        exclude_properties = set()

        if hide_readonly_properties:
            exclude_properties.update({})

        _dict = self.model_dump(serialize_as_any=True, by_alias=True, exclude=exclude_properties, exclude_none=True)

        # set to None if value (nullable) is None
        if self.value is None:
            _dict["value"] = None

        return _dict

    def to_dict_without_readonly_properties(self) -> dict[str, Any]:
        """Return the dictionary representation of the model without readonly properties."""
        return self.to_dict(hide_readonly_properties=True)

    @classmethod
    def from_dict(cls, obj: dict) -> FunctionArgument:
        """Create an instance of FunctionArgument from a dict."""
        if obj is None:
            return None

        if type(obj) is not dict:
            return FunctionArgument.model_validate(obj)

        _obj = FunctionArgument.model_validate(
            {
                "name": obj.get("name"),
                "datatype": obj.get("datatype") if obj.get("datatype") is not None else "TEXT",
                "value": obj.get("value"),
            }
        )

        return _obj


class FunctionArgumentModel:
    def __init__(
        self,  # optional properties
        name: Optional[str] = None,
        datatype: Optional[str] = "TEXT",
        value: Optional[object] = None,
    ):
        """A model object representing the FunctionArgument resource.

        Constructs an object of type FunctionArgument with the provided properties.

        Parameters
        __________
        name : str, optional
            Argument's name
        datatype : str,  default 'TEXT'
            Argument's type
        value : object, optional
            Argument's value
        """
        self.name = name
        self.datatype = datatype
        self.value = value

    __properties = ["name", "datatype", "value"]

    def __repr__(self) -> str:
        return repr(self._to_model())

    def _to_model(self):
        return FunctionArgument(
            name=self.name,
            datatype=self.datatype,
            value=self.value,
        )

    @classmethod
    def _from_model(cls, model) -> FunctionArgumentModel:
        return FunctionArgumentModel(
            name=model.name,
            datatype=model.datatype,
            value=model.value,
        )

    def to_dict(self):
        """Create a dictionary of the properties from a FunctionArgument.

        This method constructs a dictionary with the key-value entries corresponding to the properties of the FunctionArgument object.

        Returns
        _______
        dict
            A dictionary object created using the input model.
        """
        return self._to_model().to_dict()

    @classmethod
    def from_dict(cls, obj: dict) -> FunctionArgumentModel:
        """Create an instance of FunctionArgument from a dict.

        This method constructs a FunctionArgument object from a dictionary with the key-value pairs of its properties.

        Parameters
        ----------
        obj : dict
            A dictionary whose keys and values correspond to the properties of the resource object.

        Returns
        _______
        FunctionArgument
            A FunctionArgument object created using the input dictionary; this will fail if the required properties are missing.
        """
        return cls._from_model(FunctionArgument.from_dict(obj))


FunctionArgument._model_class = FunctionArgumentModel



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#
import pytest

from snowflake.core._root import Root


pytestmark = [pytest.mark.skip_gov]


def test_embed(root: Root):
    resp = root.cortex_embed_service.embed("e5-base-v2", ["foo", "bar"])

    assert len(resp.data) == 2
    assert resp.model == "e5-base-v2"
    assert resp.usage.total_tokens >= 0



from contextlib import suppress

import pytest

from snowflake.core.catalog_integration import CatalogIntegration, ObjectStore
from snowflake.core.exceptions import NotFoundError
from tests.utils import random_string

from ...utils import ensure_snowflake_version


CATALOG_INTEGRATION_PREFIX = random_string(10, "test_catalog_integration_iter_")


@pytest.fixture(scope="session")
def catalog_integrations_extended(catalog_integrations, snowflake_version):
    ensure_snowflake_version(snowflake_version, "8.37.0")

    name_list = []
    for _ in range(5):
        name_list.append(random_string(10, CATALOG_INTEGRATION_PREFIX + "a_"))
    for _ in range(7):
        name_list.append(random_string(10, CATALOG_INTEGRATION_PREFIX + "b_"))
    for _ in range(3):
        name_list.append(random_string(10, CATALOG_INTEGRATION_PREFIX + "c_"))

    for catalog_integration_name in name_list:
        catalog_integrations.create(
            CatalogIntegration(
                name=catalog_integration_name, catalog=ObjectStore(), table_format="ICEBERG", enabled=True
            )
        )

    try:
        yield catalog_integrations
    finally:
        for catalog_integration_name in name_list:
            with suppress(NotFoundError):
                catalog_integrations[catalog_integration_name].drop()


def test_iter_raw(catalog_integrations_extended):
    assert len(list(catalog_integrations_extended.iter())) >= 15


def test_iter_like(catalog_integrations_extended):
    assert len(list(catalog_integrations_extended.iter(like="test_catalog_integration_iter_"))) == 0
    assert len(list(catalog_integrations_extended.iter(like=CATALOG_INTEGRATION_PREFIX + "a_%%"))) == 5
    assert len(list(catalog_integrations_extended.iter(like=CATALOG_INTEGRATION_PREFIX + "b_%%"))) == 7
    assert len(list(catalog_integrations_extended.iter(like=CATALOG_INTEGRATION_PREFIX + "c_%%"))) == 3



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from snowflake.core.cortex.chat_service._generated.api.cortex_chat_api import CortexChatApi

__all__ = [
    "CortexChatApi",
]



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#
import json

import pytest

from urllib3 import HTTPResponse

from snowflake.core._root import Root
from snowflake.core.rest import SSEClient


pytestmark = [pytest.mark.skip_gov]


def test_cortex_inference(root: Root):
    pass
    # This test is disabled because cortex inference is not enabled on prod yet.
    # This test is manually verified to be working, see: https://github.com/snowflakedb/snowpy/pull/679
    #
    # messages = [CompleteRequestMessagesInner(content="some message")]
    # req = CompleteRequest(model="my_model", messages=messages)
    # res = root.cortex_inference_service.complete(req)
    # for e in res.events():
    #     print(e)


def test_cortex_inference_xp(root: Root):
    body_str = (
        "["
        '{"event":"model1","comment":"testing","data":{"strProp":"test1","doubleProp":1.1,"flag":false}},'
        '{"event":"model2","data":{"strProp":"test2","doubleProp":2.2,"flag":false},"retry":200},'
        '{"id":"3","event":"model3","comment":"with ID",'
        '"data":{"strProp":"test2","doubleProp":2.3,"flag":true},"retry":1000}'
        "]"
    )

    response = HTTPResponse(
        body=body_str.encode("utf-8"), status=200, headers={"Content-Type": "application/json"}, preload_content=False
    )
    res = SSEClient(response)

    event_num = 0
    for e in res.events():
        event_num += 1
        if e.event == "model1":
            assert e.comment == "testing"
        if e.event == "model2":
            assert e.retry == 200
        if e.event == "model3":
            assert e.id == "3"

    assert event_num == 3


def test_cortex_complete_xp(root: Root):
    body_str = '{"choices": [{"message": {"content": "Hello! \\n"}}]}'

    response = HTTPResponse(
        body=body_str.encode("utf-8"), status=200, headers={"Content-Type": "application/json"}, preload_content=False
    )
    res = SSEClient(response)

    for e in res.events():
        data = json.loads(e.data)
        assert data["choices"][0]["message"]["content"] == "Hello! \n"



from snowflake.core.network_rule import NetworkRule


test_network_rule_template = NetworkRule(
    name="to_be_set",
    type="HOST_PORT",
    mode="EGRESS",
    value_list=["example.com:443", "api.example.com:443"],
    comment="Test network rule",
)

test_network_rule_minimal_template = NetworkRule(
    name="to_be_set",
    type="IPv4",
)
